(*module Debug = Camlp4.Debug *)
open Printf
open Module2d
(*open Common*)

let sbprintf fmt =
  let buff = Buffer.create 42 in
  let f = Format.formatter_of_buffer buff in
  Format.kfprintf (fun f -> Format.pp_print_flush f (); Buffer.contents buff) f fmt


type exp = 
  | Fun1 of string * exp
  | Fun2 of string * exp * exp
  | Inl of exp
  | Inr of exp
  | Unit
  | Var of string
  | Let of string * exp * exp
  | Pair of exp * exp
  | MatchCase of exp * (string * exp) * (string * exp)
  | MatchPair of exp * string * string * exp

type func =
  { fname : string ; body : exp; arg1 : string; arg2 : string option }

let pp = Format.fprintf

let rec pp_exp f =
  function
  | Fun1 (s, e) ->
      pp f "%s %a" s pp_exp e
  | Fun2 (s, e1, e2) ->
      pp f"%s %a %a" s pp_exp e1 pp_exp e2
  | Inl Unit ->
      pp f "Inl()"
  | Inr Unit ->
      pp f "Inr()"
  | Inl e  ->
      pp f "Inl %a" pp_exp e
  | Inr e ->
      pp f "Inr %a" pp_exp e
  | Unit ->
      pp f "()"
  | Var s ->
      pp f "%s" s
  | Pair (e1, e2) ->
      pp f "(%a,%a)" pp_exp e1 pp_exp e2
  | Let (s, e1, e2) ->
      pp f "let %s = %a in %a" s pp_exp e1 pp_exp e2
  | MatchCase (e, (s1,e1), (s2,e2)) ->
      pp f "match %a with [ %s -> %a | %s -> %a ]"
        pp_exp e s1 pp_exp e1 s2 pp_exp e2
  | MatchPair (e1, s1, s2, e2) ->
      pp f "let (%s, %s) = %a in %a"
        s1 s2 pp_exp e2 pp_exp e1

let rec pp_exp_ast f =
  function
  | Fun1 (s, e) ->
      pp f "Fun1 %S (%a)" s pp_exp_ast e
  | Fun2 (s, e1, e2) ->
      pp f"Fun2 %S (%a) (%a)" s pp_exp_ast e1 pp_exp_ast e2
  | Inl e  ->
      pp f "Inl (%a)" pp_exp_ast e
  | Inr e ->
      pp f "Inr (%a)" pp_exp_ast e
  | Unit ->
      pp f "Unit"
  | Var s ->
      pp f "Var %S" s
  | Let (s, e1, e2) ->
      pp f "Let %S (%a) (%a)" s pp_exp_ast e1 pp_exp_ast e2
  | Pair (e1, e2) ->
      pp f "Pair (%a) (%a)" pp_exp_ast e1 pp_exp_ast e2
  | MatchCase (e, (s1,e1), (s2,e2)) ->
      pp f "MatchCase (%a) (%S, %a) (%S, %a)"
        pp_exp_ast e s1 pp_exp_ast e1 s2 pp_exp_ast e2
  | MatchPair (e1, s1, s2, e2) ->
      pp f "MatchPair (%a) %S %S (%a)"
        pp_exp_ast e1 s1 s2 pp_exp_ast e2

let pp_src f _ = pp f "<src: not_impl>";

module Fv = Set.Make(String)
module Env = struct
  include Map.Make(String)
  let find x y =
    try find x y with Not_found ->
      failwith ("unbound value: "^x)
  let pp pp_elt f e =
    pp f "{";
    iter (fun s -> pp f " %S: %a" s pp_elt) e;
    pp f " }"
end

let merge = Fv.union
let hide = Fv.remove

let rec fv = function
  | Fun1 (_,e) | Inl e | Inr e -> fv e
  | Fun2 (_,e1,e2) | Pair (e1,e2) -> merge (fv e1) (fv e2)
  | Unit -> Fv.empty
  | Var x -> Fv.singleton x
  | Let (x,e1,e2) -> abstr x e2 (fv e1)
  | MatchCase (e,(x1,e1),(x2,e2)) -> abstr x1 e1 (abstr x2 e2 (fv e))
  | MatchPair (e, x1,x2, e1) -> abstr2 x1 x2 e1 (fv e)

and abstr x e f = merge (hide x (fv e)) f
and abstr2 x1 x2 e f = merge (hide x1 (hide x2 (fv e))) f

let count = ref 0
let boxes = ref []

let mkbox lab ~n ~w () =
  let b = { label = lab; in_n = n; in_w = w } in
  boxes := b :: !boxes;
  let c = !count in
  incr count;
  c

let nth_box n =
  List.nth (List.rev !boxes) n

let box_e lab ?n ?w () = let i = mkbox lab ?n ?w () in Out_e i
let box_es lab ?n ?w () = let i = mkbox lab ?n ?w () in (Out_e i, Out_s i)

let buffer_kill = ref true
let to_kill = ref None
let box_kill f1 = 
  if !buffer_kill then
    match !to_kill with
      | Some f2 -> 
	  ignore (mkbox "send[]" ~n:(Some f1) ~w:(Some f2) ());
	  to_kill := None
      | None ->
	  to_kill := Some f1
  else
    ignore (mkbox "send[]" ~w:(Some f1) ~n:None ())

let flush_kill () =
  match !to_kill with
    | Some f -> 
	prerr_endline "FLUSH KILL";
	ignore (mkbox "send[]" ~w:(Some f) ~n:None ());
	to_kill := None
    | None -> 
	()



let split env fv1 fv2 = 
  Env.fold
    (fun x f (env1,env2) ->
       match Fv.mem x fv1, Fv.mem x fv2 with
	 | true, false -> Env.add x f env1, env2
	 | false, true -> env1, Env.add x f env2
	 | false, false -> assert false
	 | true, true ->
	     let f1,f2 = box_es "split(N,N)" ~n:f () in
	     Env.add x f1 env1, Env.add x f2 env2
    ) env (Env.empty,Env.empty)

let clean env e =
  let fve = fv e in
  Env.fold 
    (fun x f keep ->
       if Fv.mem x fve then Env.add x f keep
       else (box_kill f; keep)) 
    env Env.empty

let rec holes = function
  | Unit -> 0
  | Inl e | Inr e -> holes e
  | Pair (e1,e2) -> holes e1 + holes e2
  | _ -> 1

let top_val = function
  | Inl _ | Inr _ | Pair _ | Unit -> true
  | _ -> false

let rec gapify gaps e =
  if List.length gaps = 1 && holes e >= 2  then 
    match gaps with [hd] -> hd,[],[e] | _ -> assert false
  else match e with
  | Unit -> ("()",gaps,[])
  | Inl Unit -> ("Inl()",gaps,[])
  | Inl e -> 
      let (s,gaps,es) = gapify gaps e in 
      "Inl " ^ s, gaps, es
  | Inr Unit -> ("Inr()",gaps,[])
  | Inr e -> 
      let (s,gaps,es) = gapify gaps e in 
      "Inr " ^ s, gaps,es
  | Pair (e1,e2) ->
(*      if Random.int 2 = 0 then
	let (s1,gaps1,es1) = gapify gaps e1 in
	let (s2,gaps2,es2) = gapify gaps1 e2 in
	Printf.sprintf "(%s,%s)" s1 s, gaps2, es1@es2
      else *)
	let (s1,gaps1,es1) = gapify gaps e2 in
	let (s2,gaps2,es2) = gapify gaps1 e1 in
	Printf.sprintf "(%s,%s)" s2 s1, gaps2, es1@es2
  | e ->
      match gaps with hd::tl -> hd,tl,[e] | [] -> assert false
  

(* TODO: combine "send" (value) boxes, for non constants *)

(*
let rec is_const =
  function
  | Unit -> true
  | Inl e | Inr e -> is_const e
  | Pair (e1, e2) -> is_const e1 && is_const e2
  | _ -> false
*)

let rec gen_val env e =
  let s,gaps,es = gapify ["W";"N"] e in
(*  Printf.eprintf "send[(%s,E)]\n" s;
  List.iter (fun es -> Format.eprintf ": %a@." pp_exp_ast es) es; *)
  let es = List.map (compile env) es in
  let w = if List.length es > 0 then Some (List.nth es 0) else None in
  let n = if List.length es > 1 then Some (List.nth es 1) else None in
  box_e (Printf.sprintf "send[(%s,E)]" s) ?w ?n ()

and compile env e =
(*  let env = clean env e in *)  (* a bugge... *)
(*  debug compile "e: <<%a>>, env: %a@." pp_exp_ast e (Env.pp pp_src) env in*)
  if top_val e then gen_val env e
  else
(*  if is_const e then box_e (sbprintf "send[(%a,E)]" pp_exp e) () else *)
  match e with
  | Fun1 (f,e) -> box_e (sprintf "use %s" f) ~n:(compile env e) ()
  | Fun2 (f,e1,e2) ->
      let env1,env2 = split env (fv e1) (fv e2) in
      box_e (sprintf "use %s" f) ~n:(compile env1 e1) ~w:(compile env2 e2) ()
  | Var x -> Env.find x env
  | Let (x,e1,e2) ->
      let env1,env2 = split env (fv e1) (hide x (fv e2)) in
      compile (Env.add x (compile env1 e1) env2) e2

  | MatchCase (_e,(_x1,_e1),(_x2,_e2)) ->
      let () = Format.eprintf "MatchCase not in tail position@." in
(*      box_e "use error" () *)
      failwith "MatchCase not in tail position" 
  | MatchPair (e1,x1,x2,e2) ->
      let env1,env2 = split env (fv e1) (hide x1 (hide x2 (fv e2))) in
      let (f2,f1) = box_es "split N" ~n:(compile env1 e1) () in
      compile (Env.add x1 f1 (Env.add x2 f2 env2)) e2
  | _ -> assert false

let rec duplicate f = function
  | 0 -> box_kill f; []
  | 1 -> [f]
  | n ->
      let f1,f2 = box_es "split(N,N)" ~n:f () in
      let f1,f2 = if Random.int 2 = 0 then f2,f1 else f1,f2 in
      f1::(duplicate f2 (n - 1))

let origin = function
  | Out_e b | Out_s b -> b
  | In_n | In_w -> (-1)

let is_use b =
  let l = b.label in
  String.sub l 0 4 = "use "

let transistor sw f =
  let t () = box_e "send[(W,E)]" ~w:f ~n:sw () in
  match f with
    | Out_e b | Out_s b when origin sw < origin f ->
	let b = nth_box b in
	if is_use b then t ()
	else
	if b.in_n = None then (b.in_n <- Some sw; f)
	else if b.in_w = None then (b.in_w <- Some sw; f)
	else t ()
    | _ -> t ()

let rec compile_res env e =
  let env = clean env e in
(*  debug compile_res "e: <<%a>>, env: %a@." pp_exp_ast e (Env.pp pp_src) env in*)
  match e with
  | MatchCase (e,(x1,e1),(x2,e2)) ->
      let fv1 = hide x1 (fv e1) and fv2 = hide x2 (fv e2) in
      let env0,env12 = split env (fv e) (merge fv1 fv2) in
      let env1,env2 = split env12 fv1 fv2 in
      let (f1,f2) = box_es "case N of E,S" ~n:(compile env0 e) () in

      let branch env f x e =
	if not (Fv.mem x (fv e)) 
	then 
	  let rs = compile_res env e in
	  let fs = duplicate f (List.length rs) in
	  List.map2 transistor fs rs
	else
	compile_res (Env.add x f env) e
      in
      branch env1 f1 x1 e1 @ branch env2 f2 x2 e2 

  | Let (x,e1,e2) ->
      let env1,env2 = split env (fv e1) (hide x (fv e2)) in
      compile_res (Env.add x (compile env1 e1) env2) e2
  | MatchPair (e1,x1,x2,e2) ->
      let env1,env2 = split env (fv e1) (hide x1 (hide x2 (fv e2))) in
      let (f2,f1) = box_es "split N" ~n:(compile env1 e1) () in
      compile_res (Env.add x1 f1 (Env.add x2 f2 env2)) e2
  | e -> [compile env e]


let compile_fun name env e = 
  boxes := [];
  count := 0;
  buffer_kill := true;
  let outputs = compile_res env e in
  let outputs = 
    if !to_kill != None then (
      buffer_kill := false;
      to_kill := None;
      boxes := [];
      count := 0;
      compile_res env e
    ) else outputs in
  { name = name;
    outputs = outputs;
    boxes = Array.of_list (List.rev !boxes) }

let compile_fun1 name x e = 
  compile_fun name (Env.add x In_n Env.empty) e

let compile_fun2 name x1 x2 e = 
  compile_fun name (Env.add x1 In_n (Env.add x2 In_w Env.empty)) e


let draw_fun1 oc name x e =
  Canvas.draw_module oc (compile_fun1 name x e)
let draw_fun2 oc name x y e =
  Canvas.draw_module oc (compile_fun2 name x y e)

let draw_fun oc f =
  match f.arg2 with
    | None -> draw_fun1 oc f.fname f.arg1 f.body
    | Some arg2 -> draw_fun2 oc f.fname f.arg1 arg2 f.body
