open ExtList
open Count

module Strategy = struct

  type oracle = Prop.t list * Prop.t option * Prop.t -> nb_proofs option

  let no_oracle : oracle = fun _ -> None

  type strategy = {
    max_search_depth : int option;
    oracle: oracle;
    explore_when_unprovable : bool;
    explore_when_infinite : bool;
    just_count : bool; 
    reexplore_abandoned : bool;
  }

  (* Only the number of proofs interests us, we abandon asap
    nodes for which it is known *) 
  let counting = {
    max_search_depth = None;
    oracle = no_oracle ;
    explore_when_unprovable = false;
    explore_when_infinite = false;
    just_count = true;
    reexplore_abandoned = false;
  }

  (* This exploration is used for code generation, we explore
   known-provable nodes in order to be able to create as much
   lambda-terms as possible afterwards *)
  let explore_provable = {
    max_search_depth = None;
    oracle = no_oracle ;
    explore_when_unprovable = false;
    explore_when_infinite = true;
    just_count = false;
    reexplore_abandoned = false;}

  (* For displaying purposes : we also explore useless branches *)
  let explore_all = {
    max_search_depth = None;
    oracle = no_oracle ;
    explore_when_unprovable = true;
    explore_when_infinite = true;
    just_count = false;
    reexplore_abandoned = true;
  }

end

module type LJTSig = sig
type env_t

type problem = { env : env_t; stoup : Prop.t option; goal : Prop.t; }
val string_text_of_problem : ?initial_env:env_t * string -> problem -> string
val string_latex_of_problem : ?initial_env:env_t * string -> problem -> string
val string_of_problem : ?initial_env:env_t * string -> problem -> string

type node_rule =
  | Ax
  | Cont of Prop.t * problem
  | ImpL of problem * problem
  | ImpR of problem
  | AndL of (int * int) * problem
  | AndR of problem list
  | OrL of Prop.or_nb * problem list
  | OrR of Prop.or_nb * (int * int) * problem

type proofs_found1 = [
  | `AllFound of nb_proofs
  | `SomeFound of nb_proofs ]

type proofs_found2 = [
  | `SomeFoundOfInfiniteNb of nb_proofs
  | `SomeFoundOfFiniteNb of nb_proofs ]

type proofs_found = [ proofs_found1 | proofs_found2 ]

val string_of_proofs_found : proofs_found -> string

type node_problem = {
  mutable hypotheses : node_rule list ;
  mutable nb_proofs : proofs_found ;

(* For walking *)
  mutable nb_visits : int ;
  mutable proofs_paths : Paths.t list }

val known_problem : problem -> bool
val retrieve_problem_node : problem -> node_problem

val solve_problem : Strategy.strategy -> problem -> Count.nb_proofs * bool

val store_nodes_to_reexplore : bool ref
val reexplore : bool -> unit
end

module Build (Env : Env.EnvSig) : LJTSig with type env_t = Env.t =
struct
open Strategy

let store_nodes_to_reexplore = ref true

type env_t = Env.t
type path_t = Paths.t

type problem = {
  env : Env.t ;
  stoup : Prop.t option ;
  goal : Prop.t }

let string_text_of_problem ?initial_env pb = 
  Printf.sprintf "%s ; %s |- %s" 
    (Env.to_string_text ?initial_env pb.env)
    (match pb.stoup with None -> "" | Some p -> Prop.to_string_text p)
    (Prop.to_string_text pb.goal)

let string_latex_of_problem ?initial_env pb = 
  Printf.sprintf "%s;%s\\vdash %s" 
    (Env.to_string_latex ?initial_env pb.env)
    (match pb.stoup with None -> "\\emptystoup" 
       | Some p -> Prop.to_string_latex p)
    (Prop.to_string_latex pb.goal)

let string_of_problem ?initial_env pb =
  if !Options.result = Options.LaTeX then
    string_latex_of_problem ?initial_env pb
  else
    string_text_of_problem ?initial_env pb


type node_rule = 
  | Ax
  | Cont of Prop.t * problem
  | ImpL of problem * problem
  | ImpR of problem
  | AndL of (int * int) * problem
  | AndR of problem list
  | OrL of Prop.or_nb * problem list
  | OrR of Prop.or_nb * (int * int) * problem

type proofs_found1 = [
  | `AllFound of nb_proofs
  | `SomeFound of nb_proofs ]

type proofs_found2 = [
  | `SomeFoundOfInfiniteNb of nb_proofs
  | `SomeFoundOfFiniteNb of nb_proofs ]

type proofs_found = [ proofs_found1 | proofs_found2 ]

let string_of_proofs_found (p : proofs_found) =
  if !Options.result = Options.LaTeX then 
    match p with
      | `AllFound n -> to_string_nb_proofs n
      | `SomeFound n -> Printf.sprintf "\\ge%s" (to_string_nb_proofs n)
      | `SomeFoundOfInfiniteNb n -> 
          Printf.sprintf "%s/\\infty" (to_string_nb_proofs n)
      | `SomeFoundOfFiniteNb n ->
          Printf.sprintf "%s/<\\infty" (to_string_nb_proofs n)
  else
    match p with
      | `AllFound n -> to_string_nb_proofs n
      | `SomeFound n -> Printf.sprintf ">=%s" (to_string_nb_proofs n)
      | `SomeFoundOfInfiniteNb n -> 
          Printf.sprintf "%s/infty" (to_string_nb_proofs n)
      | `SomeFoundOfFiniteNb n ->
          Printf.sprintf "%s/<infty" (to_string_nb_proofs n)


type node_problem = {
  mutable hypotheses : node_rule list ;
  mutable nb_proofs : proofs_found ;
  mutable nb_visits : int ;
  mutable proofs_paths : Paths.t list }
  
let nodes_problem : (problem, node_problem) Hashtbl.t = Hashtbl.create 57

let known_problem = Hashtbl.mem nodes_problem

let register_problem pb =
  let node = { hypotheses = [] ; nb_proofs = `SomeFound Zero ;
               nb_visits = 0 ; proofs_paths = []} in
    Hashtbl.add nodes_problem pb node ;
    node

let retrieve_problem_node = Hashtbl.find nodes_problem

let apply_rules pb =
  match pb.stoup with
    | Some p -> (
        match Prop.prop_of_t p with
          | Prop.Arrow (a, b) ->
              [ImpL ({ env = pb.env ; stoup = None ; goal = a},
                     { env = pb.env ; stoup = Some b ; goal = pb.goal })]
          | Prop.And l -> let s = List.length l in 
              List.mapi (fun i p -> AndL ((i, s), { pb with stoup = Some p})) l
          | Prop.Or (i, l) -> [OrL (i, List.map 
                                     (fun p -> { env = Env.add_elem pb.env p ; 
                                           stoup = None ; goal = pb.goal }) l)]
          | Prop.Var _ -> if p = pb.goal then [Ax] else [])
    | None ->
        (match Prop.prop_of_t pb.goal with
           | Prop.Arrow (a, b) -> 
               [ImpR { env = Env.add_elem pb.env a ; stoup = None ; goal = b }]
           | Prop.And l -> [AndR (List.map (fun p -> { pb with goal = p }) l)]
           | Prop.Or (i, l) -> let s = List.length l in
               List.mapi (fun i' p -> OrR (i, (i', s), { pb with goal = p })) l
           | Prop.Var _ -> [])@
        (List.map 
           (fun p -> Cont (p, { pb with stoup = Some p }))
           (Env.elems_multiset pb.env))
    
let stop_solving strategy depth =
  match strategy.max_search_depth with
    | None -> false
    | Some v -> v < depth

module PbsSet = Set.Make(struct type t = problem let compare = compare end)

module CountState = Count.CountState(struct 
                                       type t = problem 
                                       let compare = compare end)
open CountState

exception NoProof
exception InfiniteNumberOfProofs

let (add_abandoned_depth, reset_abandoned_depth, abandoned_depth) =
  let l = ref [] in
    (fun p -> l := p :: !l),
    (fun () -> l := []),
    (fun p -> List.mem p !l)

let to_reexplore = ref []

let store_node_to_reexplore pb =
  if !store_nodes_to_reexplore then
    to_reexplore := pb :: !to_reexplore

let max_proofs_found (n1 : proofs_found) (n2 : proofs_found1) : proofs_found =
  match n1, n2 with
    | `AllFound v1, `AllFound v2 -> assert (v1 = v2) ; n1
    | `AllFound _, `SomeFound _ -> assert false;

    | `SomeFound v1, `SomeFound v2 -> `SomeFound (max_nb_proofs v1 v2)
    | `SomeFound v1, `AllFound v2 -> assert (v2 >=% v1) ; (n2 :> proofs_found)

    | `SomeFoundOfInfiniteNb _, `AllFound (Zero | Finite _) -> assert false
    | `SomeFoundOfInfiniteNb _, `AllFound Infinite -> (n2 :> proofs_found)
    | `SomeFoundOfInfiniteNb _, `SomeFound Zero -> n1
    | `SomeFoundOfInfiniteNb _, `SomeFound Infinite -> 
        `SomeFoundOfInfiniteNb Infinite
    | `SomeFoundOfInfiniteNb Infinite, _ -> n1
    | `SomeFoundOfInfiniteNb Zero, `SomeFound (Finite _ as v) ->
        `SomeFoundOfInfiniteNb v
    | `SomeFoundOfInfiniteNb (Finite v1), `SomeFound (Finite v2) ->
        `SomeFoundOfInfiniteNb (Finite (Big_int.max_big_int v1 v2))

    | `SomeFoundOfFiniteNb _, `AllFound (Zero | Infinite) -> assert false
    | `SomeFoundOfFiniteNb _, `AllFound (Finite _) -> (n2 :> proofs_found)
    | `SomeFoundOfFiniteNb _, `SomeFound Zero -> n1
    | `SomeFoundOfFiniteNb _, `SomeFound Infinite -> assert false
    | `SomeFoundOfFiniteNb Infinite, _ -> assert false
    | `SomeFoundOfFiniteNb Zero, `SomeFound (Finite _ as v) ->
        `SomeFoundOfFiniteNb v
    | `SomeFoundOfFiniteNb (Finite v1), `SomeFound (Finite v2) ->
        `SomeFoundOfFiniteNb (Finite (max v1 v2))


let update_nb_proofs node v =
  node.nb_proofs <- max_proofs_found node.nb_proofs v

let all_found n all_found =
  if all_found then `AllFound n else `SomeFound n

let use_oracle oracle pb node =
  if oracle != Strategy.no_oracle then
    match node.nb_proofs with
      | #proofs_found2 -> () (* An oracle has already been called *)
      | `AllFound _ -> () (* All proofs have already been found *)
      | `SomeFound v -> 
          match oracle (Env.elems_set pb.env, pb.stoup, pb.goal) with
            | None -> ()
            | Some Zero -> node.nb_proofs <- `AllFound Zero
            | Some Infinite -> node.nb_proofs <- `SomeFoundOfInfiniteNb v
            | Some (Finite _)-> node.nb_proofs <- `SomeFoundOfFiniteNb v

let (set_debug, ppb, pres) =
  let debug = ref false in
  let step = ref 0 in
    (fun b -> debug := b),
    (fun pb ->
       if !debug then (
         incr step;
         prerr_string (String.make (2* !step) ' ');
         prerr_string (string_text_of_problem pb);
         prerr_newline ())),
    (fun reason r ->
       if !debug then (
         prerr_string (String.make (2* !step) ' ');
         prerr_string (Count.to_string_nb_proofs r.nb) ; 
         prerr_string " ";
         prerr_string reason;
(*         List.iter (fun e -> Printf.eprintf " %s ." (string_text_of_problem e))
           (elements r.loop);*)
         prerr_newline ();
         decr step);
       r)

(* Note : r is true iif lpbs if different from the empty set *)
let rec solve_problem strategy depth lpbs pb =
  ppb pb;
  let solve_node depth node =
    use_oracle strategy.oracle pb node;
    match node.nb_proofs, strategy.just_count, 
      strategy.reexplore_abandoned with
      | `AllFound v, _, _ ->
          pres "AlreadyFound" { nb = v ; explored = true ; loop = no_loop}
      | (`SomeFound v | `SomeFoundOfFiniteNb v | `SomeFoundOfInfiniteNb v),
        _, false when abandoned_depth pb ->
          pres "Abandoned" { nb = v ; explored = false ; loop = no_loop }
      | `SomeFoundOfInfiniteNb _, true, _ ->
          pres "Inf found" { nb = Infinite ; explored = false ; loop = no_loop}
      | _ -> 
          if PbsSet.mem pb lpbs then (* Found a loop, we stop here *)
            pres "Loop" { nb = Zero ; explored = true ; loop = singleton pb }
          else
            let r = solve_node_rule_list strategy depth 
                      (PbsSet.add pb lpbs) node.hypotheses in
            let loop' = remove_loop r.loop pb in
              if contains_singleton r.loop pb then (
                let n = if r.nb <> Zero then Infinite else Zero in
                  update_nb_proofs node (all_found n 
                                           (loop'=no_loop && r.explored));
                  pres "Singl" { loop = loop' ; nb = n ; 
                                 explored = r.explored})
              else (
                if r.loop <> no_loop then (
                  store_node_to_reexplore pb;
                  update_nb_proofs node (`SomeFound r.nb);
                  pres "NormalLoop" { r with loop = loop'})
                else (
                  update_nb_proofs node (all_found r.nb r.explored);
                  pres "NormalNoLoop" {r with loop = loop'}))
  in
    try 
      let node = retrieve_problem_node pb in
        solve_node depth node
    with Not_found -> (* The problem was not previously encountered *)
      if stop_solving strategy depth then (
        add_abandoned_depth pb;
        pres "Abandoning" { nb = Zero ; explored = false ; loop = no_loop})
      else
        let node = register_problem pb in 
        let rules = apply_rules pb in
          node.hypotheses <- rules;
          solve_node (depth + 1) node

and solve_problem_list strategy depth lpbsroot l = 
  let rec aux r = function
    | [] -> r
    | t::q  -> 
        let r' = solve_problem strategy depth lpbsroot t in
          if r' = unprovable then (
            if strategy.explore_when_unprovable then List.iter
              (fun pb -> ignore (solve_problem strategy depth lpbsroot pb)) q;
            raise NoProof)
          else
            aux (r *$ r') q
  in
    try aux count_one l
    with NoProof -> count_zero

and solve_node_rule strategy depth lpbs = function
  | Ax -> count_one
  | Cont (_, p) | ImpR p | AndL (_, p) | OrR (_, _, p) -> 
      solve_problem strategy depth lpbs p
  | ImpL (p1, p2) -> solve_problem_list strategy depth lpbs [p1 ; p2]
  | AndR l | OrL (_, l) -> solve_problem_list strategy depth lpbs l

and solve_node_rule_list strategy depth lpbsroot l =
  let rec aux r = function
    | [] -> r
    | t::q  -> 
        let r' = solve_node_rule strategy depth lpbsroot t in
          if r'.nb = Infinite && strategy.explore_when_infinite = false then
            raise InfiniteNumberOfProofs
          else
            aux (r +$ r') q
  in 
    try aux count_zero l
    with InfiniteNumberOfProofs -> { count_infinity with explored = false }

let solve_problem strategy pb =
  let r = solve_problem strategy 0 PbsSet.empty pb in
    assert (r.loop = no_loop);
    r.nb, r.explored

let reexplore verbose =
  let already_reexplored = ref PbsSet.empty in
    while !to_reexplore <> [] do
      let l = !to_reexplore in
        to_reexplore := [];
        List.iter (fun pb ->
                   if PbsSet.mem pb !already_reexplored = false then (
                     already_reexplored := PbsSet.add pb !already_reexplored;
                     let (nb, exp)= solve_problem Strategy.explore_all pb in
                       if verbose then Printf.eprintf "%s: %s proofs\n"
                       (string_text_of_problem pb) (to_string_nb_proofs nb))) l
  done
end
