open Options

type pb = 
  | NoneChosen
  | Builtin of int
  | Text of string

let pb = ref NoneChosen

let display_grammar = ref false

let speclist = [
  ("--set", Arg.Unit (fun _ -> solver := Set), "Solves in LJT^Set");
  ("--mset", Arg.Unit (fun _ -> solver := MSet), "Solves in LJT^MSet");

  ("--inf_search_depth", Arg.Unit (fun _ -> depth_search := None),
   "Search will go at an arbitrary depth");
  ("--fin_search_depth", Arg.Int (fun i -> depth_search := Some i),
   "Goes no deeper than the specified depth search");

  ("--use_oracle", Arg.Set use_oracle, 
   "Use an oracle for solving in LJT^MSet");
  ("--inf_oracle_depth", Arg.Unit (fun _ -> depth_oracle := None),
   "The oracle will go at an arbitrary depth");
  ("--fin_oracle_depth", Arg.Int (fun i -> depth_oracle := Some i),
   "The oracle won't go no deeper than the specified depth search");
  
  ("--explore_count", Arg.Unit (fun _ -> explore_type := Counting),
   "During search, stops as soon as an infinite is encountered");
  ("--explore_enumerate", Arg.Unit (fun _ -> explore_type := TermEnumeration),
   "During search, explores all branches which could lead to the finding of a proof");
  ("--explore_all", Arg.Unit (fun _ -> explore_type := ExploreAll),
   "During search, explores even useless branches. Not guaranteed to be correct");

  ("--nb_proofs", Arg.Unit (fun _ -> output := NbProofs),
   "Returns only the number of proofs of the problem");
  ("--terms", Arg.Unit (fun _ -> output := CamlCode),
   "Returns ocaml program fragments corresponding to the proofs");
  ("--parseable", Arg.Unit (fun _ -> output := Parseable),
   "Returns parseable data corresponding to the proofs");
  ("--graph_text", Arg.Unit (fun _ -> output := GraphText),
   "Returns GSet or GMSet, with text anotations");
  ("--graph_latex", Arg.Unit (fun _ -> output := GraphLaTeX),
   "Returns GSet or GMSet, with LaTeX anotations");

  ("--unroll", Arg.Int (fun i -> walk_depth := i),
   "Unroll cyclic proofs n times (for terms generation)");

  ("--pb_builtin", Arg.Int (fun i -> pb := Builtin i),
   "Solves the ith builtin problem");
  ("--pb", Arg.String (fun s -> pb := Text s),
   "Solves the given problem");

  ("--grammar", Arg.Set display_grammar,
   "Display the grammar of accepted problems")
]

open Input

let grammar =
"You must specify a valid problem, i.e. which follows the following grammar
(where pb is the terminal symbol):

int    ::= ['1'-'9']['0'-'9']*
TYPE   ::= 't'int
LBRACE ::= '('
RBRACE ::= ')'
AND    ::= '&' | \"/\\\" | '*'
OR     ::= '|' | \"\\/\"
AROW   ::= \"->\" 
SEP    ::= ','
VDASH  ::= \"|-\"

pb ::= env VDASH prop

env ::=
| 
| fenv

fenv:
| prop 
| prop SEP env

prop:
| TYPE
| LBRACE prop RBRACE
| prop AND prop
| prop OR  prop
| prop ARROW prop

AND or OR have the same priority, and bind more tightly than ARROW.
"

let _ =
  Arg.parse speclist (fun _ -> ()) "";
  if !display_grammar then (print_string grammar; exit 0);
  let pb = match !pb with
    | NoneChosen -> 
        prerr_string  
          "No problem specified, use --pb or --pb_builtin to specify one\n";
        exit 1
    | Builtin i ->
        (try 
           let env, goal = List.nth Input.lpb i in
             env, None, goal
         with (Failure "nth") -> 
           prerr_string  "Not enough builtin problems\n";
           exit 1)
    | Text s ->
        try 
          Parser_prop.pb Lexer_prop.token (Lexing.from_string s)
        with 
          | Parsing.Parse_error ->
              prerr_string "Failure: Parse error.\n";
              prerr_string grammar;
              exit 1
          | Failure "lexing: empty token" -> 
              prerr_string "Failure: Lexing error.\n";
              prerr_string grammar;
              exit 1
  in
    match !solver with
      | Set -> Main.main_set (convert_pb_set pb)
      | MSet -> Main.main_mset (convert_pb_mset pb)
