open ExtList

module type EnvSig = sig
  type t

  val compare : t -> t -> int

  val empty : t

  val elems_set : t -> Prop.t list
  val elems_multiset : t -> Prop.t list

  val add_elem : t -> Prop.t -> t
  val add_list_elem : t -> Prop.t list -> t
  val of_list : Prop.t list -> t
  val member : t -> Prop.t -> bool
    
  val to_string_text : ?annot:(Prop.t -> int -> string -> string) -> 
    ?initial_env:(t * string) -> t -> string
  val to_string_latex : ?annot:(Prop.t -> int -> string -> string) -> 
    ?initial_env:(t * string) -> t -> string
  val to_string : ?annot:(Prop.t -> int -> string -> string) ->
    ?initial_env:(t * string) -> t -> string
end

module EnvSet : EnvSig = struct

type t = int
let compare (v1 : t) (v2 : t) = v1 - v2

external int_of_elem : Prop.t -> int = "%identity"
external eleml_of_intl : int list -> Prop.t list = "%identity"

type env = Ptset.t

let envs : (t, env) Hashtbl.t = Hashtbl.create 57 
let ts : (env, t) Hashtbl.t = Hashtbl.create 57

let _ = 
  Hashtbl.add envs 0 Ptset.empty;
  Hashtbl.add ts Ptset.empty 0
let empty = 0

let next_t : t ref = ref 1

let retrieve_or_create (p : env) =
  try Hashtbl.find ts p
  with Not_found ->
    let t = !next_t in
      Hashtbl.add ts p t;
      Hashtbl.add envs t p;
      incr next_t;
      t

let retrieve_from_t = Hashtbl.find envs

let elems_set t = eleml_of_intl (Ptset.elements (retrieve_from_t t))
let elems_multiset = elems_set

let add_elem t (e : Prop.t) = 
  let s = Hashtbl.find envs t in
    retrieve_or_create (Ptset.add (int_of_elem e) s)

let add_list_elem = List.fold_left add_elem

let of_list = add_list_elem empty

let member t (e : Prop.t) =
  let s = Hashtbl.find envs t in
    Ptset.mem (int_of_elem e) s

let to_string f_prop ?(annot = fun _ _ x -> x) ?initial_env e = 
  let flist, e = match initial_env with
    | Some (0, _) (* empty initial env *) 
    | None -> (fun l -> l), retrieve_from_t e
    | Some (initial, name) -> (fun l -> name :: l),
      Ptset.diff (retrieve_from_t e) (retrieve_from_t initial)
  in
    String.concat ", " (flist (List.map (fun p -> annot p 0 (f_prop p)) (eleml_of_intl (Ptset.elements e))))

let to_string_text = to_string Prop.to_string_text
let to_string_latex = to_string Prop.to_string_latex
let to_string ?annot ?initial_env p = 
  if !Options.result = Options.LaTeX then 
    to_string_latex ?annot ?initial_env p
  else
    to_string_text ?annot ?initial_env p

end

module EnvMultiSet : EnvSig = struct

type t = int
let compare (v1 : t) (v2 : t) = v1 - v2

external prop_of_int : int -> Prop.t = "%identity"
external int_of_prop : Prop.t -> int = "%identity"

type env = int Ptmap.t

let envs : (t, env) Hashtbl.t = Hashtbl.create 57 
let ts : (env, t) Hashtbl.t = Hashtbl.create 57

let _ = 
  Hashtbl.add envs 0 Ptmap.empty;
  Hashtbl.add ts Ptmap.empty 0
let empty = 0

let next_t : t ref = ref 1

let retrieve_or_create (p : env) =
  try Hashtbl.find ts p
  with Not_found ->
    let t = !next_t in
      Hashtbl.add ts p t;
      Hashtbl.add envs t p;
      incr next_t;
      t

(* Takes an env, and returns a set list of Prop.t *)
let elems_set t = 
  Ptmap.fold (fun e _ l -> prop_of_int e :: l) (Hashtbl.find envs t) []


(* Takes an env, and returns a multiset list of Prop.t *)
let elems_multiset t = List.flatten (
  Ptmap.fold (fun e card l -> (List.init card (fun _ -> prop_of_int e)) :: l)
             (Hashtbl.find envs t) [])


let add_elem t (e : Prop.t) = 
  let s = Hashtbl.find envs t in
  let card =
    try Ptmap.find (int_of_prop e) s
    with Not_found -> 0
  in
    retrieve_or_create (Ptmap.add (int_of_prop e) (card + 1) s)

let add_list_elem = List.fold_left add_elem

let of_list = add_list_elem empty

let member t e =
  let s = Hashtbl.find envs t in
    Ptmap.mem (int_of_prop e) s

(* "Substract" two lists *)
let rec alter_list alter = function
  | [] -> []
  | (key, value)::q ->
      let value' = value - try List.assoc key alter with Not_found -> 0 in
        if value' <= 0 then alter_list alter q
        else (key, value') :: alter_list alter q

(* Ad Hoc function : returns a couple (prop, card) *)
let elems_card t = Ptmap.fold (fun e card l -> (prop_of_int e, card) :: l) 
                (Hashtbl.find envs t) []


let id_annot (x : Prop.t) (i : int) (s : string) = s

(* If the user specifies the optional function for printing, we use
   the long style format. Otherwise we factorize all elements of the
   same types *)
let to_string f_prop ?(annot = id_annot) ?initial_env e =
  let flist, e = match initial_env with
    | Some (0, _) (* empty initial env *)
    | None -> (fun l -> l), elems_card e
    | Some (initial, name) -> 
        (fun l -> name :: l), 
        alter_list (elems_card initial) (elems_card e) in
    String.concat ", "
      (flist (if annot == id_annot then 
                List.map (fun (p, i) -> if i = 1 then f_prop p else
                            Printf.sprintf "%s [%d]" (f_prop p) i) e
              else
                List.flatten (List.map (fun (p, i) -> let t = f_prop p in
                   List.init i (fun i -> annot p i t)) e)))

let to_string_text = to_string Prop.to_string_text
let to_string_latex = to_string Prop.to_string_latex
let to_string ?annot ?initial_env p = 
  if !Options.result = Options.LaTeX then 
    to_string_latex ?annot ?initial_env p
  else
    to_string_text ?annot ?initial_env p

end
