module Build (LJTMS : LJT.LJTSig with type env_t = Env.EnvMultiSet.t) : sig
  exception BadPath
  val interpret_proof :
    LJTMS.problem -> EnvLambda.env_lambda -> Paths.t -> LambdaBar.t

end =
struct

open LJTMS
open Paths
open LambdaBar
open ExtList

exception BadPath

let rec find_proof path lproofs = 
  let path' = path_of_t path in
    match (lproofs, path') with
      | [], _ -> raise BadPath
      | Ax::_, PAx -> Ax, path'
      | ImpL _ as p :: q, PImpL _ -> p, path'
      | ImpR _ as p :: q, PImpR _ -> p, path'
      | AndR _ as p :: q, PAndR _ -> p, path'
      | AndL (i1, _) as p :: q, PAndL (i2, _) ->
          if i1 = i2 then p, path' else find_proof path q
      | Cont (e1, _) as p :: q, PCont ((e2, _), _) -> 
          if e1 = e2 then p, path' else find_proof path q
      | OrL _ as p :: q, POrL _ -> p, path'
      | OrR (_, i1, _) as p :: q, POrR (_, i2, _) ->
          if i1 = i2 then p, path' else find_proof path q
      | _::q, _ -> find_proof path q

let rec interpret_proof pb env path =
  try
    let node = retrieve_problem_node pb in
    let proof = find_proof path node.hypotheses in 
      match proof with
        | Ax , PAx -> empty_list
        | ImpL (pb1, pb2), PImpL (pa1, pa2) ->
            let u = extract_term (interpret_proof pb1 env pa1)
            and l = extract_arguments (interpret_proof pb2 env pa2) in
              cons u l
        | Cont (p, pb), PCont ((_, i), path) ->
            let v = EnvLambda.retrieve_term env p i
            and l = extract_arguments (interpret_proof pb env path) in
              apply v l
        | ImpR pb', PImpR pa ->
            let (a, b) = match Prop.prop_of_t pb.goal with
                Prop.Arrow (a, b) -> (a, b) | _ -> assert false in
            let va = new_var () in
            let env' = EnvLambda.add_interpretation env a va  in
            let x = extract_term (interpret_proof pb' env' pa) in
              abs va x
        | AndL (i, pb), PAndL (_, pa) ->
            let l = extract_arguments (interpret_proof pb env pa) in
              conspi i l
        | AndR lpb, PAndR lpa ->
            let lx = List.map2 
                       (fun pb pa -> extract_term (interpret_proof pb env pa))
                       lpb lpa
            in
              uple lx
        | OrL (i, lpb), POrL (_, lpa) ->
            let s = match pb.stoup with Some s -> s | None -> assert false in
            let dis = match Prop.prop_of_t s with
                Prop.Or (_, l) -> l | _ -> assert false in
            let lpb' = List.map2 (fun pb p -> new_var (), pb, p) lpb dis in
            let l = List.map2 (fun (v, pb, p) pa -> 
                v, extract_term (interpret_proof pb 
                           (EnvLambda.add_interpretation env p v) pa)) lpb' lpa
            in
              case i l
        | OrR (i, i', pb), POrR (_, _, pa) -> 
            let t = extract_term (interpret_proof pb env pa) in
              inj i (fst i') t
        | _, _ -> assert false
  with Not_found -> raise BadPath
end
