open ExtList

type var = 
  | Named of (Options.result -> string)
  | Auto of int

let print_var = function
  | Named v -> v !Options.result
  | Auto i -> 
      if !Options.result = Options.LaTeX then
        Printf.sprintf "x_%d" i
      else
        Printf.sprintf "x%d" i

type term = 
  | Var of var 
  | App of var * arguments
  | Abs of var * term
  | Uple of term list
  | Inj of int * int * term
and arguments = 
  | NoArg
  | ArgT of term * arguments
  | Pi of (int * int) * arguments
  | Case of int * (var * term) list

type t = 
  | BasicVar of var
  | Term of term 
  | Arguments of arguments

let rec normalize_term = function
  | Var _ as v -> v
  | Abs (v, t) -> Abs (v, normalize_term t)
  | App (v, NoArg) -> Var v
  | App (v, l) -> App (v, normalize_arguments l)
  | Uple l -> Uple (List.map normalize_term l)
  | Inj (i, i', t) -> Inj (i, i', normalize_term t)
and normalize_arguments = function
  | NoArg -> NoArg
  | Case (i, l) -> Case (i, List.map (fun (v, t) -> (v, normalize_term t)) l)
  | Pi (v, q) -> Pi (v, normalize_arguments q)
  | ArgT (t, q) -> ArgT (normalize_term t, normalize_arguments q)

let conditional_normalize v =
  if !Options.normalize_lambda_terms = false then v
  else match v with
    | BasicVar _ as v -> v
    | Term t -> Term (normalize_term t)
    | Arguments a -> Arguments (normalize_arguments a)

let empty_list = Arguments NoArg
let apply v a = conditional_normalize (Term (App (v, a)))
let var v = BasicVar v
let abs v t = conditional_normalize (Term (Abs (v, t)))
let cons t l = conditional_normalize (Arguments (ArgT (t, l)))
let uple l = Term (Uple l)
let inj i i' t = Term (Inj (i, i', t))
let conspi (i, s) l = Arguments (Pi ((i, s), l))
let case i c = Arguments (Case (i, c))

let extract_term = function
  | Term t -> t
  | BasicVar _ | Arguments _ -> failwith "extract_term"

let extract_arguments = function
  | Arguments l -> l
  | Term _ | BasicVar _ -> failwith "extract_arguments"

let ref_var = ref 0

let new_var () = 
  let r = !ref_var in
    incr ref_var;
    Auto r

let reset_var v = ref_var := v

let namedvar v = Named v

type lterm =
  | LVar of var
  | LAbs of var * lterm
  | LApp of lterm * lterm
  | LPi of (int * int) * lterm
  | LUple of lterm list
  | LInj of int * int * lterm
  | LCase of int * lterm * (var * lterm) list

let rec lterm_of_term = function
  | Var v -> LVar v
  | App (v, l) -> lterm_of_app (LVar v) l
  | Abs (v, t) -> LAbs (v, lterm_of_term t)
  | Uple l -> LUple (List.map lterm_of_term l)
  | Inj (i, i', t) -> LInj (i, i', lterm_of_term t)
and lterm_of_app t = function
  | NoArg -> t
  | ArgT (t', q) -> lterm_of_app (LApp (t, lterm_of_term t')) q
  | Pi ((n, s), q) -> lterm_of_app (LPi ((n, s), t)) q
  | Case (i, c) -> LCase (i, t, List.map (fun (v, t) -> v, lterm_of_term t) c)

let joker = Named (function
                     | Options.LaTeX -> "\\_\\_"
                     | _ -> "__")

let lterm_of_t = function
  | BasicVar v -> LVar v
  | Term t -> lterm_of_term t
  | Arguments l -> lterm_of_term (Abs (joker, (App (joker, l))))

let rec string_of_lterm = function
  | LVar v -> print_var v
  | LAbs (v, t) -> Printf.sprintf "(fun %s -> %s)" 
      (print_var v) (string_of_lterm t)
  | LApp (t1, t2) -> Printf.sprintf "(%s %s)" 
      (string_of_lterm t1) (string_of_lterm t2)
  | LPi ((n, s), t) -> 
      Printf.sprintf "(let (%s) = %s in x)"
       (String.concat ", " (List.init s (fun i -> if i = n then "x" else "_")))
        (string_of_lterm t)
  | LUple l -> Printf.sprintf "(%s)"
      (String.concat ", " (List.map string_of_lterm l))
  | LInj (i, i', t) ->
      Printf.sprintf "`C%d_%d (%s)" i i' (string_of_lterm t)
  | LCase (i, t, l) -> Printf.sprintf "(match (%s) with %s)"
      (string_of_lterm t) (String.concat " | "
        (List.mapi (fun i' (v, t) -> Printf.sprintf "`C%d_%d %s -> %s" 
                      i i' (print_var v) (string_of_lterm t)) l))

let rec string_of_lterm_parseable = function
  | LVar v -> print_var v
  | LAbs (v, t) -> Printf.sprintf "(FUN %s %s)" 
      (print_var v) (string_of_lterm t)
  | LApp (t1, t2) -> Printf.sprintf "(APPLY %s %s)" 
      (string_of_lterm t1) (string_of_lterm t2)
  | LPi ((n, s), t) -> 
      Printf.sprintf "(let (%s) = %s in x)"
       (String.concat ", " (List.init s (fun i -> if i = n then "x" else "_")))
        (string_of_lterm t)
  | LUple l -> Printf.sprintf "(%s)"
      (String.concat ", " (List.map string_of_lterm l))
  | LInj (i, i', t) ->
      Printf.sprintf "`C%d_%d (%s)" i i' (string_of_lterm t)
  | LCase (i, t, l) -> Printf.sprintf "(match (%s) with %s)"
      (string_of_lterm t) (String.concat " | "
        (List.mapi (fun i' (v, t) -> Printf.sprintf "`C%d_%d %s -> %s" 
                      i i' (print_var v) (string_of_lterm t)) l))
