(* Main *)

let sf = Printf.sprintf;;

open Syntax;;

let string_of_token = function
| EOF -> "EOF"
| MISSING -> "MISSING"
| AND -> "AND"
| BROKEN -> "BROKEN"
| IT -> "IT"
| IS -> "IS"
| A -> "A"
| COLON -> "COLON"
| LPAREN -> "LPAREN"
| RPAREN -> "RPAREN"
| COMMA -> "COMMA"
| PROMPT -> "PROMPT"
| THE -> "THE"
| DOT -> "DOT"
| L_BROKEN_R -> "L_BROKEN_R"
| BLA u -> sf "BLA(%S)" u
| PART x -> sf "PART(%S)" x
| COLOR d -> sf "COLOR(%d)" d
| CONDITION -> "CONDITION"
| PRISTINE -> "PRISTINE"
| IN -> "IN"
;;

let dump_lexemes fn =
  let ic = open_in fn in
  let lexbuf = Lexing.from_channel ic in
  let i = ref 0 in
  let continue = ref true in
  while !continue do
    let token = Lexic.token lexbuf in
    Printf.printf "%-5d %s\n%!" !i (string_of_token token);
    incr i;
    if token = EOF then continue := false
  done
;;


let parse fn =
  let ic = open_in fn in
  let lexbuf = Lexing.from_channel ic in
  Syntax.prose Lexic.token lexbuf
;;

open Parts;;

(*let parse_applyomatic_output ic =
  let rex1 = Pcre.regexp ~study:true "^\s*val v(\\d+)" : 
  while true do
  done*)

module DS = Set.Make(struct type t = description let compare = compare end);;

let process ?(verbose=true) p =
  let parts = Hashtbl.create 256 in
  let part_id = ref 0 in
  let number_id = ref 0 in
  let number_to_id = Hashtbl.create 1024 in
  let id_to_number = Hashtbl.create 1024 in
  let id_of_number id = Hashtbl.find id_to_number id in
  let number_of_id x =
    try
      Hashtbl.find number_to_id x
    with
    | Not_found ->
        let id = !number_id in
        incr number_id;
        Hashtbl.add id_to_number id x;
        Hashtbl.add number_to_id x id;
        id
  in
  List.iter
    begin function
      | Blah -> ()
      | Part(number, description, color) ->
          let id = number_of_id number in
          (*if Hashtbl.mem parts (id, color) then
            begin
              Printf.printf "WARNING: Part %s color %s is not unique, ignoring\n%!"
                            number
                            Lexic.colors.(color)
            end
          else
            begin*)
              Hashtbl.add parts id (!part_id, color, description);
              incr part_id
    end
    p;
  if verbose then Printf.printf "Total %d parts and %d numbers.\n%!" (Hashtbl.length parts) (Hashtbl.length id_to_number);
  let bf = Printf.bprintf in
  let b = Buffer.create 256 in
  let string_of_description ?(numeric=true) id p =
    let rec loop = function
    | Number n ->
        if numeric then
          bf b "t%d" (number_of_id n)
        else
          bf b "%s" n
    | Missing(p1,p2) ->
        bf b "(";
        loop p1;
        bf b " -> ";
        loop p2;
        bf b ")";
    | And(p1,p2) ->
        bf b "(";
        loop p1;
        bf b " /\\ ";
        loop p2;
        bf b ")"
    | Pristine ->
        if numeric then
          bf b "t%d" id
        else
          bf b "%s" (id_of_number id) 
    in
    Buffer.clear b;
    loop p;
    Buffer.contents b
  in
  let dump ?(numeric=false) ~verbose () =
    let virgule = ref false in
    let l = ref [] in
    Hashtbl.iter
      begin fun k v ->
        l := (k,v) :: !l
      end
      parts;
    let a = Array.of_list !l in
    Array.sort begin fun (_,(x,_,_)) (_,(y,_,_)) -> compare x y end a;
    Array.iter
      begin fun (id, (part_id, color, description)) ->
        if verbose then
          Printf.printf " Part #%03d is of type %S (#%d) and of color %S: %s\n"
                        part_id
                        (id_of_number id)
                        id
                        Lexic.colors.(color)
                        (string_of_description ~numeric id description)
        else
          begin
            if !virgule then Printf.printf ", ";
            virgule := true;
            Printf.printf "%s" (string_of_description ~numeric id description)
          end
      end
      a;
    Printf.printf "\n"
  in
  dump ~verbose:true ();
  dump ~numeric:true ~verbose:false ();
  let produced = Hashtbl.create 16 in
  let being_produced = Hashtbl.create 16 in
  let rec produce' active = function
    | Pristine ->
        true
    | And(d1,d2) ->
        produce active d1;
        produce active d2;
        true
    | Missing(d1,d2) ->
        produce active d1;
        produce active d2;
        true
    | Number n ->
        let t = number_of_id n in
        (* Get rules *)
        try
          let candidates = Hashtbl.find_all parts t in
          List.exists
            begin fun (part_id, color, description) ->
              produce active description
            end
            candidates
        with
        | Not_found -> false
  and produce active d = 
    try
      Hashtbl.find produced d
    with
    | Not_found ->
      if DS.mem d active then
        false
      else
        let p = produce' (DS.add d active) d in
        Hashtbl.add produced d p;
        p
  in
  let n = Sys.argv.(2) in
  let ok = produce DS.empty (Number n) in
  Printf.printf "Part %s is %s.\n%!"
    n
    (if ok then "produceable" else "not produceable.")
;;

let _ =
  let fn = Sys.argv.(1) in
  (*dump_lexemes fn;*)
  let p = parse fn in
  process p
;;
