(* Blacksearch *)

exception Invalid_move;;

module IM = Map.Make(struct type t = int let compare = compare end);;

type problem = {
  pb_sigma : int array;
  pb_rho : int array
}

type state = (int * int) IM.t;; (* position -> marble, number of plonks for that marble *)
type move = int;; (* Swap at given position *)

let bf = Printf.bprintf;;
let sf = Printf.sprintf;;

let string_of_state pb q =
  let b = Buffer.create 256 in
  let m = Array.length pb.pb_sigma in
  bf b "(";
  for i = 0 to m - 1 do
    let (x,_) = IM.find i q in 
    if i > 0 then bf b "-";
    bf b "%d" x
  done;
  bf b "):(";
  for i = 0 to m - 1 do
    let (_,p) = IM.find i q in 
    if i > 0 then bf b "-";
    bf b "%d" p
  done;
  bf b ")";
  Buffer.contents b
;;

type metric = {
  mt_permutations_to_go : int;
  mt_plonks_to_go : int
};;

let string_of_metric m = sf "(%d,%d)" m.mt_permutations_to_go m.mt_plonks_to_go;;

type configuration = {
  cf_parent : (configuration * move) option;
  cf_state : state;
  cf_metric : metric;
};;

module QS = Set.Make(struct type t = state let compare = compare end);;

(*** apply_move *)
let apply_move pb q m x =
  (* Swap x and x + 1 *)
  let y = x + 1 in
  let (mx,px) = IM.find x q
  and (my,py) = IM.find y q
  in
  if px = 0 then
    raise Invalid_move
  else
    begin
      let delta q i =
        let (m,p) = IM.find i q in (* Find the marble at position i *)
        if i <> pb.pb_sigma.(m) then 1 else 0
      in
      let q' = IM.add x (my,py) (IM.add y (mx,px - 1) q) in
      (q',
       { mt_plonks_to_go = m.mt_plonks_to_go - 1;
         mt_permutations_to_go = m.mt_permutations_to_go - (delta q x) - (delta q y) + (delta q' x) + (delta q' y) })
    end
;;
(* ***)
let iota m = Array.init m (fun i -> i);;
let swap a i j =
  let x = a.(i) in
  a.(i) <- a.(j);
  a.(j) <- x
;;
let randomize a =
  let m = Array.length a in
  for i = 0 to m - 2  do
    let j = i + Random.int (m - i) in
    swap a i j
  done
;;

module type GENOME =
  sig
    val randomize_moves : bool
    val downwards_moves : bool
    val permutations_first : bool
    val permutations_downwards : bool
    val plonks_downwards : bool
    val memory_forget_percentage : int
  end
;;

module G1 =
  struct
    let randomize_moves = false
    let downwards_moves = false
    let permutations_first = true
    let permutations_downwards = false
    let plonks_downwards = true
    let memory_forget_percentage = 30
  end
;;

module G2 =
  struct
    let randomize_moves = false
    let downwards_moves = true
    let permutations_first = true
    let permutations_downwards = true
    let plonks_downwards = true
    let memory_forget_percentage = 30
  end
;;

module G3 =
  struct
    let randomize_moves = false
    let downwards_moves = false
    let permutations_first = true
    let permutations_downwards = true
    let plonks_downwards = true
    let memory_forget_percentage = 30
  end
;;
(*** Metric *)

let ( *** ) x y = if x = 0 then y else x ;;

module Solver(G : GENOME) =
  struct
    (*** iterate_over_moves *)
    let iterate_over_moves pb f =
      let m = Array.length pb.pb_sigma in
      for i = 0 to m - 2 do
        f i
      done
    ;;
    (* ***)

    module Metric =
      struct
        type t = metric;;

        let compare m1 m2 =
          (compare m1.mt_permutations_to_go m2.mt_permutations_to_go) ***
          (compare m1.mt_plonks_to_go m2.mt_plonks_to_go)
        ;;
      end
    ;;

    (*** metric_of_state *)
    let metric_of_state pb q =
      let permutations_to_go = ref 0
      and plonks_to_go = ref 0
      in
      let m = Array.length pb.pb_sigma in
      for i = 0 to m - 1 do
        let (m,p) = IM.find i q in (* Find the marble at position i *)
        if i <> pb.pb_sigma.(m) then incr permutations_to_go;
        plonks_to_go := !plonks_to_go + p
      done;
      { mt_permutations_to_go = !permutations_to_go;
        mt_plonks_to_go = !plonks_to_go }
    ;;
    (* ***)

    (*** is_promising *)
    let is_promising pb q =
      let m = Array.length pb.pb_sigma in
      let rec loop i =
        if i = m then
          true
        else
          let (x,p) = IM.find i q in
          (* Marble x is at position i *)
          (* It needs to be at position sigma.(x) *)
          let y = pb.pb_sigma.(x) in
          (y <= i or y - i <= p)
          (*(p = 0 or
            for_all_over_range
              (i + 1)
              (y - 1)
              (fun j -> let (x',p') = IM.find j q in p' > 0))*)
      in
      loop 0
    ;;
    (* ***)
    (*** is_final *)
    let is_final pb c =
      let m = c.cf_metric in
      m.mt_permutations_to_go = 0 && m.mt_plonks_to_go = 0
    ;;
    (* ***)
    (*** rec *)
    let rec for_all_over_range i0 i1 f =
      if i0 > i1 then
        true
      else
        f i0 && (for_all_over_range (i0 + 1) i1 f)
    ;;
    (* ***)

    module PQ = Prioqueue.Make(Metric);;

    (*** initial_configuration *)
    let initial_configuration pb =
      let q = ref IM.empty in
      let m = Array.length pb.pb_sigma in
      for i = 0 to m - 1 do
        q := IM.add i (i, pb.pb_rho.(i)) !q
      done;
      { cf_state = !q;
        cf_parent = None;
        cf_metric = metric_of_state pb !q
      }
    ;;
    (* ***)
    (*** recompute_solution *)
    let recompute_solution pb c =
      let rec loop = function
        | None -> []
        | Some(c', m) -> m :: (loop c'.cf_parent)
      in
      loop c.cf_parent
    ;;
    (* ***)
    (*** solver *)
    class solver ?(debug=false) pb =
      let sigma = pb.pb_sigma
      and rho = pb.pb_rho
      in
      let m = Array.length sigma in
      object(self)
        val breadth = PQ.create ()
        val mutable memory = QS.empty
        val mutable last_metric = None
        method deja_vu q =
          if QS.mem q memory then
            begin
              true
            end
          else
            begin
              (*if G.memory_forget_percentage = 0
                 or Random.int 100 < G.memory_forget_percentage then*)
                  memory <- QS.add q memory;
              (*else
                ();*)
              false
            end
        method add c =
          ignore (PQ.add breadth c c.cf_metric)
        initializer
          let c0 = initial_configuration pb in
          self#add c0
        method get_status =
          match last_metric with
          | None -> ""
          | Some m -> string_of_metric m
        method step =
          if PQ.is_empty breadth then
            raise Not_found
          else
            let cpq = PQ.take breadth in
            let m = cpq.PQ.priority in
            last_metric <- Some m;
            let c = cpq.PQ.contents in
            if debug then Format.printf "Popping configuration %s with metric %s.@\n" (string_of_state pb c.cf_state) (string_of_metric m);
            if is_final pb c then
              Some (recompute_solution pb c)
            else
              begin
                self#children c;
                None
              end
        method children c =
          let q = c.cf_state in
          if debug then Format.printf "Checking children: @[";
          iterate_over_moves pb
            begin fun mv ->
              try
                let (q',m') = apply_move pb q c.cf_metric mv in
                if debug then Format.printf "  Move %d gives %s.@\n" m (string_of_state pb q');
                if is_promising pb q' then
                  begin
                    if not (self#deja_vu q') then
                      begin
                        let c = { cf_state = q';
                                  cf_parent = Some(c, m);
                                  cf_metric = m' }
                        in
                        self#add c
                      end
                    else
                      begin
                        if debug then Format.printf "  State already seen.@\n"
                      end
                  end
                else
                  if debug then Format.printf "  Move not promising.@\n"
              with
              | Invalid_move ->
                  if debug then Format.printf "  Move %d invalid.@\n" m
            end;
          if debug then Format.printf "@]@.";
      end
    ;;
    (* ***)
end
