(* Cozbakim *)

type movement  = Down | Right | Left;;
type dance = movement array;;

module IS = Set.Make(struct type t = int let compare = compare end);;

(*** inverse *)
let inverse a =
  let m = Array.length a in
  let b = Array.make m 0 in
  for i = 0 to m - 1 do
    b.(a.(i)) <- i
  done;
  b
;;
(* ***)
(*** load *)
let load fn =
  let rex = Pcre.regexp ~study:true "^(\\d+) -> \\((\\d+),(\\d+)\\)$" in
  let ic = open_in fn in
  let res = ref [] in
  let max_x = ref 0 in
  try
    while true do
      let l = input_line ic in
      try
        let g x = int_of_string (Pcre.get_substring (Pcre.exec ~rex l) x) in
        let (x,y,z) = (g 1, g 2, g 3) in
        res := (x,y,z) :: !res;
        max_x := max x !max_x;
      with
      | Not_found -> ()
    done;
    assert false
  with
  | End_of_file ->
      close_in ic;
      let m = 1 + !max_x in
      let (marble_to_destination, plonks) = (Array.make m 0, Array.make m 0) in
      List.iter
        begin fun (x,y,z) ->
          marble_to_destination.(x) <- y;
          plonks.(x) <- z
        end
        !res;
      (marble_to_destination, plonks)
;;
(* ***)
(*** cyclic_analysis *)
let cyclic_analysis marble_to_destination =
  let m = Array.length marble_to_destination in
  let to_do = ref IS.empty in
  for i = 0 to m - 1 do
    to_do := IS.add i !to_do
  done;
  let result = ref [] in
  let kaldir x = to_do := IS.remove x !to_do in
  while not (IS.is_empty !to_do) do
    let x = IS.min_elt !to_do in
    let rec loop l y =
      if IS.mem y !to_do then
        begin
          kaldir y;
          loop (y :: l) marble_to_destination.(y)
        end
      else
        l
    in
    if marble_to_destination.(x) = x then
      kaldir x
    else
      let cycle = List.rev (loop [] x) in
      result := cycle :: !result
  done;
  !result
;;
(* ***)
(*** movement_to_string *)
let movement_to_string h =
  let m = Array.length h in
  let u = String.create m in
  for i = 0 to m - 1 do
    u.[i] <-
      match h.(i) with
      | Down -> '|'
      | Right -> '>'
      | Left -> '<'
  done;
  u
;;
(* ***)
(*** put_x *)
let put_x m k =
  Array.init m
    begin fun i ->
      if i = k then
        Right
      else if i = k + 1 then
        Left
      else
        Down
    end
;;
(* ***)
(*** exchange *)
let rec exchange m (x,y) =
  if x = y then []
  else
    if x > y then exchange m (y,x)
    else 
      begin
        let result = ref [] in
        let xat k = result := (put_x m k) :: !result in
        let k = y - x in
        for j = 0 to k - 1 do
          xat (x + j)
        done;
        for j = k - 2 downto 0 do
          xat (x + j)
        done;
        !result
      end
;;
(* ***)
(*** compute_marble_to_position_and_price *)
let compute_marble_to_position_and_price m ?how_much result =
  let price = Array.make m 0 in
  let marble_to_position = Array.init m (fun i -> i) in
  let how_much =
    match how_much with
    | None -> Array.length result
    | Some k -> k
  in
  for k = 0 to how_much - 1 do
    let h = result.(k) in
    for i = 0 to m - 1 do
      let j = marble_to_position.(i) in
      if h.(j) = Right then price.(i) <- 1 + price.(i);
      match h.(j) with
      | Down -> ()
      | Left -> marble_to_position.(i) <- j - 1
      | Right -> marble_to_position.(i) <- j + 1
    done
  done;
  (marble_to_position, price)
;;
(* ***)
(*** print_dance *)
let print_dance result =
  Array.iter
    begin fun h ->
      Printf.printf "%s\n" (movement_to_string h)
    end
    result
;;
(* ***)
(*** print_residuals *)
let print_residuals marble_to_destination (marble_to_position,result,price,plonks) =
  let m = Array.length price in
  for i = 0 to m - 1 do
    Printf.printf "%d\n" (plonks.(i) - price.(i))
  done
;;
(* ***)
(*** print *)
let print marble_to_destination (marble_to_position,result,price,plonks) =
  print_dance result;
  let m = Array.length price in
  for i = 0 to m - 1 do
    let c =
      match compare price.(i) plonks.(i) with
      | -1 -> '-'
      | 0  -> ' '
      | _ -> '+'
    in
    Printf.printf " %c %-3d/%-3d (remaining %d)\n" c price.(i) plonks.(i) (plonks.(i) - price.(i))
  done;
  for i = 0 to m - 1 do
    if marble_to_position.(i) <> marble_to_destination.(i) then
      begin
        Printf.printf "Hole #%d goes to #%d and not to #%d !!!\n%!"
          i
          marble_to_position.(i)
          marble_to_destination.(i)
      end
  done;
  Printf.printf "\n%!"
;;
(* ***)
(*** cycle_to_exchanges *)
let cycle_to_exchanges p =
  let sort (x,y) = if x < y then (x,y) else (y,x) in
  let result = match p with
  | [] -> []
  | [_] -> []
  | x :: y :: rest ->
      let rec loop head = function
      | y :: z :: rest -> (sort (y,z)) :: (loop head (z :: rest))
      | [z] -> []
      | [] -> []
      in
      (sort (x,y)) :: (loop x (y :: rest))
  in
  List.rev result
;;
(* ***)
(*** remove_duplicates *)
let rec remove_duplicates = function
  | [] -> []
  | [x] -> [x]
  | x :: y :: rest ->
      if x = y then
        rest
      else
        x :: (remove_duplicates (y :: rest))
;;
(* ***)
(*** azalt *)
let azalt (result,price,plonks) =
  let m = Array.length price in
  let result' = ref result in
  let okay = ref true in
  while !okay do
    okay := false;
    for i = 0 to m - 2 do
      if price.(i) + 2 <= plonks.(i) && price.(i + 1) + 2 <= plonks.(i + 1) then
        begin
          result' := (exchange m (i,i+1)) @ !result';
          price.(i) <- price.(i) + 2;
          okay := true;
        end
    done
  done;
  (!result',price,plonks)
;;
(* ***)
(*** cycle_decomposition_solution *)
let cycle_decomposition_solution (marble_to_destination, plonks) =
  let m = Array.length marble_to_destination in
  let cyclic = cyclic_analysis marble_to_destination in
  let result =
      (List.concat
        (List.map
          begin fun cycle ->
            let dtl = cycle_to_exchanges cycle in
            List.concat (List.map (exchange m) dtl)
          end
          cyclic))
  in
  let result = remove_duplicates result in
  let result = Array.of_list result in
  let (marble_to_position,price) = compute_marble_to_position_and_price m result in
  (marble_to_position, result, price, plonks)
;;
(* ***)
let iota m = Array.init m (fun i -> i);;
let swap a i j =
  let x = a.(i) in
  a.(i) <- a.(j);
  a.(j) <- x
;;
(*** alt_result *)
let alt_result (marble_to_destination, plonks) =
  let m = Array.length marble_to_destination in
  let marble_to_destination' = inverse marble_to_destination in
  let result = ref [] in
  let marble_to_position = iota m in
  let rec loop i =
    if i = m - 1 then
      ()
    else
      begin
        let j = m - 1 - i in
        let x = marble_to_destination'.(j) in (* Le # du marbre qui doit arriver a la derniere position *)
        let position_to_marble = inverse marble_to_position in
        let z = position_to_marble.(x) in
        (*Printf.printf "Iteration #%d: le marbre #%d doit arriver en #%d alors qu'il est en #%d.\n%!"
                      i
                      x
                      j
                      z;*)
        (* Deplacer le marbre x de sa position (marble_to_position.(x))
         * jusqu'a sa position finale j *)
        for k = z to j - 1 do
          result := (put_x m k) :: !result;
          (*Printf.printf " (%d,%d)\n%!" k (k + 1);*)
          swap marble_to_position k (k + 1)
        done;
        loop (i + 1)
      end
  in
  loop 0;
  let result = Array.of_list (List.rev !result) in
  let (marble_to_position,price) = compute_marble_to_position_and_price m result in
  (marble_to_position, result, price, plonks)
;;
(* ***)
(*** alt_result2 *)
let alt_result2 (marble_to_destination, plonks) =
  let m = Array.length marble_to_destination in
  let marble_to_destination' = inverse marble_to_destination in
  let result = ref [] in
  let marble_to_position = iota m in
  let rec loop i =
    if i = m - 1 then
      ()
    else
      begin
        let j = i in
        let x = marble_to_destination'.(j) in (* Le # du marbre qui doit arriver a la premiere position *)
        let position_to_marble = inverse marble_to_position in
        let z = position_to_marble.(x) in
        (*Printf.printf "Iteration #%d: le marbre #%d doit arriver en #%d alors qu'il est en #%d.\n%!"
                      i
                      x
                      j
                      z;*)
        (* Deplacer le marbre x de sa position (marble_to_position.(x))
         * jusqu'a sa position finale j *)
        for k = j - 1 downto z do
          result := (put_x m k) :: !result;
          (*Printf.printf " (%d,%d)\n%!" k (k + 1);*)
          swap marble_to_position k (k + 1)
        done;
        loop (i + 1)
      end
  in
  loop 0;
  let result = Array.of_list (List.rev !result) in
  let (marble_to_position,price) = compute_marble_to_position_and_price m result in
  (marble_to_position, result, price, plonks)
;;
(* ***)

module Opt =
  struct
    let problem = ref "../b000";;
    let debug = ref false;;
  end
;;

(*** GUI *)
open Graphics;;

let green = rgb 40 180 60;;

let sf = Printf.sprintf;;

let rec for_all_over_range i0 i1 f =
  if i0 > i1 then
    true
  else
    f i0 && (for_all_over_range (i0 + 1) i1 f)
;;

class display ?(x0=50) ?(y0=60) ?(p=16) marble_to_destination (marble_to_position,result,price,plonks) () =
  let grid0 = result in
  let m0 = Array.length grid0 in
  let n = Array.length grid0.(0) in
  object(self)
    val mutable grid = grid0
    val mutable m = m0
    val mutable size = min m0 40
    val mutable top = 0
    val mutable finger_i = 0
    val mutable finger_j = 0
    val mutable marble_to_position = [||]
    val mutable price = [||]
    val mutable price_mode : [`remaining|`price|`plonks] = `remaining
    method change_price_mode x = price_mode <- x
    initializer
      self#recompute
    method yatay_git j = finger_j <- j
    method kaydet fn =
      let oc = open_out fn in
      for i = 0 to m - 1 do
        Printf.fprintf oc "%s\n" (movement_to_string grid.(i))
      done;
      close_out oc
    method sil =
      for i = 0 to m - 1 do
        for j = 0 to n - 1 do
          grid.(i).(j) <- Down
        done
      done;
      self#recompute
    method recompute =
      let (d,f) = compute_marble_to_position_and_price n grid in
      marble_to_position <- d;
      price <- f
    method left = if finger_j > 0 then finger_j <- finger_j - 1 else ()
    method right = if finger_j + 1 < n then finger_j <- finger_j + 1 else ()
    method yukari =
      if finger_i > 0 then finger_i <- finger_i - 1
      else if top > 0 then top <- top - 1
      else ()
    method down =
      if finger_i < size - 1 then finger_i <- finger_i + 1
      else if top + size < m then top <- top + 1
      else ()
    method change with_what =
      let i = finger_i + top
      and j = finger_j in
      let okay =
        match with_what with
        | Right -> j + 1 < n
        | Left -> j > 0
        | Down -> true
      in
      if okay then
        begin
          grid.(i).(j) <- with_what;
          self#recompute
        end
      else
        ()
    method open_line =
      let i = top + finger_i in
      let upper = Array.sub grid 0 i in
      let lower = Array.sub grid i (m - i) in
      grid <- Array.concat [upper; [| Array.make n Down |]; lower];
      m <- m + 1;
      size <- min m 40;
      self#recompute
    method find_best_trapezoid =
      let best_trapezoid = ref None
      and best_best_trapezoid_price = ref 0
      in
      let add_trapezoid dir f i j n =
        Printf.printf "Feasible %s from %d to %d repeating %d, %d plonks\n%!"
          (if dir then "right" else "left")
          i j n f;
        if f > !best_best_trapezoid_price then
          begin
            best_trapezoid := Some(dir, i, j, n);
            best_best_trapezoid_price := f
          end
      in
      let position_to_marble = inverse marble_to_position in
      let remaining i = plonks.(position_to_marble.(i)) - price.(position_to_marble.(i)) in
      for i = 0 to n - 1 do
        Printf.printf "Marble at #%d is %d, has %d plonks left.\n" i position_to_marble.(i) (remaining i);
        for j = i + 1 to n - 1 do
          let range_min = ref max_int in
          for k = i to j do
            range_min := min !range_min (remaining k)
          done;
          let rmin = !range_min in
          let k = j - i in
          Printf.printf "%d plonks between %d and %d\n%!" rmin i j;
          if rmin = 0 then
            ()
          else
            begin
              (* Les marbres entre les positions i et j
               * ont tous au moins un plonk. *)
              Printf.printf "Marble at #%d has %d plonks left, at #%d has %d plonks left.\n%!"
                i (remaining i)
                j (remaining j);
              if remaining i >= k then
                (* Le marbre qui est actuellement en position i
                 * a au moins k plonks de plus que le marbre
                 * actuellement en position j. *)
                (* On peut donc faire un trapeze vers la droite. *)
                (* Ca va nous eliminer :
                  * - n + k plonks de i
                  * - n plonks de j
                  * - 1 plonk de chacun des i + 1 ... j - 1
                  * - soit n + k + n + j - 1 - i + 1 = 2n + k + j - i plonks *)
                begin
                  let n' =
                    min
                      (remaining i - k)
                      (remaining j)
                  in
                  if n' >= 0 then add_trapezoid true (2 * n' + k + j - i) i j n'
                end;
              (* De la droite vers la gauche ? *)
              if (*remaining i +*) remaining j >= k then
                (* Le marbre qui est actuellement en position j
                 * a au moins k plonks de plus que le marbre
                 * actuellement en position i. *)
                begin
                  let n' =
                    min
                      (remaining i)
                      (remaining j - k)
                  in
                  if n' >= 0 then add_trapezoid false (2 * n' + k + j - i) i j n'
                end;
            end
        done
      done;
      match !best_trapezoid with
      | None -> raise Not_found
      | Some(dir,i,j,n) -> (dir,i,j,n,!best_best_trapezoid_price)
    method put_x =
      self#change Right;
      self#right;
      self#change Left;
      self#left
    method open_line_above =
      let i = max 0 (top + finger_i + 1) in
      let i = min 0 (m - 1) in
      let upper = Array.sub grid 0 i in
      let lower = Array.sub grid i (m - i) in
      grid <- Array.concat [upper; [| Array.make n Down |]; lower];
      m <- m + 1;
      size <- min m 40;
      self#recompute
    method delete_line =
      let i = top + finger_i in
      let upper = Array.sub grid 0 i in
      let lower = Array.sub grid (i + 1) (m - i - 1) in
      grid <- Array.concat [upper; lower];
      m <- m - 1;
      if i = m then finger_i <- finger_i - 1;
      size <- min m 40;
      self#recompute
    method draw =
      set_color white;
      fill_rect (x0 - p) (y0 - p) ((n + 1) * p) ((size + 5) * p);
      for i = 0 to size - 1 do
        for j = 0 to n - 2 do
          let x = x0 + j * p
          and y = y0 + (size - 1 - (i - 2)) * p
          in
          if i = finger_i then
            begin
              set_color (rgb 255 255 100);
              fill_rect x y p p
            end;
          set_line_width 0;
          set_color (rgb 240 240 240);
          draw_rect x y p p;
        done
      done;
      set_color (rgb 120 120 240);
      set_line_width 2;
      draw_rect x0 y0 ((n - 1) * p) ((size + 2) * p);
      for i = 0 to size - 1 do
        for j = 0 to n - 1 do
          let x = x0 + j * p
          and y = y0 + (size - 1 - (i - 3)) * p
          in
          begin
            set_color
              begin
                if i = finger_i && j = finger_j then
                  red
                else
                  black
              end;
            set_line_width 2;
            match grid.(top + i).(j) with
            | Down ->
              moveto x y;
              lineto x (y - p)
            | Left ->
              moveto x y;
              lineto (x - p) (y - p)
            | Right ->
              moveto x y;
              lineto (x + p) (y - p)
          end;
        done
      done;
      (*let x = x0 + finger_j * p
      and y = y0 + (size - 1 - (finger_i - 2)) * p
      in
      set_color red;
      moveto x y;
      lineto x (y + p);*)
      let (marble_to_position,_) = compute_marble_to_position_and_price n ~how_much:(top + finger_i + 1) grid in
      let position_to_marble = inverse marble_to_position in
      for k = 0 to n - 1 do
        let j = position_to_marble.(k) in
        set_color
          begin
            if marble_to_destination.(j) = k then
              green
            else
              red
          end;
        moveto (x0 + k * p) (y0 + p);
        let x =
          match price_mode with
          | `remaining -> plonks.(j) - price.(j)
          | `plonks -> plonks.(j)
          | `price -> price.(j)
        in
        draw_string (sf "%d" x);
        moveto (x0 + k * p) (y0);
        draw_string (sf "%d" j)
      done;
      let finger_marble = position_to_marble.(finger_j) in
      let j = marble_to_destination.(finger_marble) in
      (*let d' = inverse d in
      let delik = d'.(finger_j) in
      let j = marble_to_destination.(delik) in*)
      set_color black;
      moveto (x0 + j * p) y0;
      lineto (x0 + j * p) (y0 + p);
  end
;;

let (width, height) = (1024,1024);;

let message msg =
  set_color blue;
  fill_rect 0 0 width 40;
  moveto 10 10;
  set_color white;
  draw_string msg
;;

let sf = Printf.sprintf;;

let gui () =
  let problem = load !Opt.problem in
  let (marble_to_destination,plonks) = problem in
  let ((marble_to_position,result,price,plonks) as whatever) = (* cycle_decomposition_solution *) alt_result problem in
  print marble_to_destination whatever;
  let quit = ref false in
  open_graph (sf " %dx%d" width height);
  auto_synchronize false;
  let g = new display marble_to_destination whatever () in
  let integer = ref 0 in
  while not !quit do
    let st = wait_next_event [Key_pressed] in
    begin
      match st.key with
      | '0'..'9' as c ->
          let d = (Char.code c) - 48 in
          integer := (10 * !integer) + d;
          message (sf "%d" !integer)
      | '\008' ->
          integer := !integer / 10;
          message (sf "%d" !integer)
      | '\013' ->
          integer := 0;
          message (sf "%d" !integer)
      | 'q' -> quit := true
      | 'k' -> g#yukari
      | 'j' -> g#down
      | 'h' -> g#left
      | 'l' -> g#right
      | '<' -> g#change Left
      | '>' -> g#change Right
      | '|' -> g#change Down
      | 'o' -> g#open_line
      | 'O' -> g#open_line_above
      | 'D' -> g#delete_line
      | '!' -> g#change_price_mode `plonks; message "Displaying required ticks"
      | '@' -> g#change_price_mode `remaining; message "Displaying remaining ticks"
      | '#' -> g#change_price_mode `price; message "Displaying actual ticks"
      | 's' -> g#kaydet (!Opt.problem^".left")
      | '\003' -> g#sil
      | 'J' ->
          for i = 1 to !integer do
            g#put_x;
            g#down
          done;
          integer := 0
      | 'H' ->
          for i = 1 to !integer do
            g#put_x;
            if i < !integer then
              begin
                g#down;
                g#left
              end
          done;
          integer := 0
      | 'L' ->
          for i = 1 to !integer do
            g#put_x;
            if i < !integer then
              begin
                g#down;
                g#right
              end
          done;
          integer := 0
      | 'x' -> g#put_x
      | 'b' ->
          begin
            try
              let (dir,i,j,n,price) as trapezoid = g#find_best_trapezoid in
              message (sf "Right=%b i=%d j=%d cost=%d..." dir i j price);
              let build_trapezoid (dir,i,j,n,_) =
                if dir then
                  begin
                    g#yatay_git i;
                    for k = 1 to 2 * (j - i + 1) + 2 * n - 2 do
                      g#open_line
                    done;
                    for k = i to j - 1 do
                      g#put_x;
                      if k < j - 1 then g#right;
                      g#down;
                    done;
                    for k = 1 to 2 * n - 2 do
                      g#put_x;
                      g#down;
                    done;
                    for k = i to j - 1 do
                      if k > i then g#left;
                      g#put_x;
                      g#down;
                    done
                  end
                else
                  begin
                    g#yatay_git (j - 1);
                    for k = 1 to 2 * (j - i + 1) + 2 * n - 2 do
                      g#open_line
                    done;
                    for k = i to j - 1 do
                      g#put_x;
                      if k < j - 1 then g#left;
                      g#down;
                    done;
                    for k = 1 to 2 * n - 2 do
                      g#put_x;
                      g#down;
                    done;
                    for k = i to j - 1 do
                      if k > i then g#right;
                      g#put_x;
                      g#down;
                    done
                  end
              in
              build_trapezoid trapezoid;
            with
            | Not_found -> message "Not found."
          end
      | ' ' ->
          g#change Down;
          g#right;
          g#change Down;
          g#left;
      | c -> message (sf "Unknown key %C" c)
    end;
    g#draw;
    synchronize ()
  done;
  close_graph ();
;;
(* ***)

open Blacksearch;;

exception Found;;

let simple () =
  let ((marble_to_destination,plonks_to_do) as problem) = load !Opt.problem in
  let ((marble_to_position,result,price,plonks) as whatever) = alt_result2 problem in
  print marble_to_destination whatever
;;

let residuals () =
  let ((marble_to_destination,plonks_to_do) as problem) = load !Opt.problem in
  let ((marble_to_position,result,price,plonks) as whatever) = alt_result2 problem in
  print_residuals marble_to_destination whatever
;;

let iter_over_trapezoids marble_to_position plonks price add_trapezoid =
  let position_to_marble = inverse marble_to_position in
  let remaining i = plonks.(position_to_marble.(i)) - price.(position_to_marble.(i)) in
  let n = Array.length marble_to_position in
  for i = 0 to n - 1 do
    for j = i + 1 to n - 1 do
      let range_min = ref max_int in
      for k = i to j do
        range_min := min !range_min (remaining k)
      done;
      let rmin = !range_min in
      let k = j - i in
      if rmin = 0 then
        ()
      else
        begin
          if remaining i >= k then
            begin
              let n' = min (remaining i - k) (remaining j) in
              if n' >= 0 then add_trapezoid true i j n'
            end;
          if remaining j >= k then
            begin
              let n' = min (remaining i) (remaining j - k) in
              if n' >= 0 then add_trapezoid false i j n'
            end;
        end
    done
  done
;;

let trapezoids () =
  let ((marble_to_destination,plonks_to_do) as problem) = load !Opt.problem in
  let ((marble_to_position,result,price,plonks) as whatever) = alt_result2 problem in
  iter_over_trapezoids marble_to_position plonks price
    begin fun dir i j n -> Printf.printf "%b %d %d %d\n" dir i j n end
;;


module Intprog =
  struct
    module IM = Map.Make(struct type t = int let compare = compare end);;
    module IS = Set.Make(struct type t = int let compare = compare end);;

    exception Undefined;;
    exception Not_satisfied;;
    exception Found of int IM.t;;

    let dump oc environment =
      Printf.printf "Solution:\n";
      IM.iter
        begin fun variable_id coefficient ->
          Printf.fprintf oc " #%d %d\n" variable_id coefficient
        end
        environment;
      Printf.printf "%!"
    ;;

    let check oc clauses solution =
      let get vi =
        try
          IM.find vi solution
        with
        | Not_found -> 0
      in
      let m = Array.length clauses in
      for i = 0 to m - 1 do
        let (objective, terms) = clauses.(i) in
        let sum = ref 0 in
        List.iter
          begin fun (variable_id, coefficient) ->
            sum := !sum + (get variable_id) * coefficient;
          end
          terms;
        if !sum <> objective then
          begin
            Printf.fprintf oc "Clause #%d not satisfied, wants %d, got %d.\n%!" i objective !sum
          end
      done
    ;;


    let solve clauses =
      let m = Array.length clauses in
      Printf.printf "Total %d clauses.\n%!" m;
      let rec solve level environment active_clauses pinned_variables =
        Printf.printf "%08d: Active %d clauses, %d vars pinned.\n"
          level
          (IS.cardinal active_clauses)
          (IS.cardinal pinned_variables);
        if IS.is_empty active_clauses then raise (Found environment);
        let get vi =
          try
            IM.find vi environment
          with
          | Not_found -> 0
        in
        let increment vi =
          assert (not (IS.mem vi pinned_variables));
          IM.add vi (1 + get vi) environment
        in
        (* Check that there are no unsatisfied clauses *)
        let satisfied_clauses = ref IS.empty in
        let compute_clause_sum i =
          let (objective, terms) = clauses.(i) in
          let sum = ref 0 in
          let undefined = ref IS.empty in
          List.iter
            begin fun (variable_id, coefficient) ->
              sum := !sum + (get variable_id) * coefficient;
              if !sum > objective then
                begin
                  Printf.printf "Clause #%d not satisfied.\n%!" i;
                  raise Not_satisfied
                end
            end
            terms;
          (!sum, objective, !undefined)
        in
        IS.iter
          begin fun ci ->
            let (sum, objective, undefined) = compute_clause_sum ci in
            if sum = objective then
              begin
                (* This clause is now satisfied.  Wonderful. *)
                Printf.printf "Clause #%d satisfied.\n%!" ci;
                satisfied_clauses := IS.add ci !satisfied_clauses
              end
          end
          active_clauses;
        (* Check if we have satisfied clauses *)
        if IS.is_empty !satisfied_clauses then
          begin
            (* Select a clause *)
            let ci = IS.choose active_clauses in
            let (sum, objective, undefined) = compute_clause_sum ci in
            (* *)
            let (objective, terms) = clauses.(ci) in
            (* Compute current sum *)
            if sum < objective then
              let delta = objective - sum in
              let terms' = Array.of_list (List.filter (fun (variable_id, coefficient) ->
                not (IS.mem variable_id pinned_variables) && coefficient <= delta) terms) in
              (* Choose best term here XXX *)
              Array.sort
                begin fun (_, coefficient1) (_,coefficient2) -> compare (delta - coefficient1) (delta - coefficient2) end
                terms';
              for i = 0 to Array.length terms' - 1 do
                let (variable_id, coefficient) = terms'.(i) in
                try
                  Printf.printf "%08d: Trying variable #%d for clause #%d.\n" level variable_id i;
                  solve (level + 1) (increment variable_id) active_clauses pinned_variables
                with
                | Not_satisfied ->
                    Printf.printf "Backtracking.\n%!"
              done
            else
              Printf.printf "Warning, clause %d...\n%!" ci
          end
        else
          begin
            (* Some clauses are satisfied, recurse. *)
            (* Pin variables *)
            let pinned_variables =
              IS.fold
                begin fun ci pinned_variables ->
                  let (objective, terms) = clauses.(ci) in
                  List.fold_left
                    begin fun pinned_variables (variable_id, _) ->
                      IS.add variable_id pinned_variables
                    end
                    pinned_variables
                    terms
                end
                !satisfied_clauses
                pinned_variables
            in
            solve level environment (IS.diff active_clauses !satisfied_clauses) pinned_variables
          end
      in
      try
        solve 0 IM.empty
          (let rec loop i =
            if i = m then
              IS.empty
            else
              IS.add i (loop (i + 1))
          in
          loop 0)
          IS.empty;
        raise Not_satisfied
      with
      | Found solution -> solution
    ;;
  end
;;

let intprogify ?(fn=(!Opt.problem)) () =
  let ((marble_to_destination,plonks_to_do) as problem) = load fn in
  let ((marble_to_position,result,price,plonks) as whatever) = alt_result2 problem in
  let trapezoids = ref [] in
  let n = Array.length marble_to_destination in
  (* Compute all trapezoids. *)
  let trapezoid_id = ref 0 in
  iter_over_trapezoids marble_to_position plonks price
    begin fun dir i j n ->
      let k = j - i in
      if dir then
        begin
          (* De la gauche vers la droite. *) 
          let rec loop l =
            if l > j then
              []
            else
              (if l = i then
                (l,k+n)
              else
                if l = j then
                  (l,n)
                else
                  (l,1)) :: (loop (l + 1))
          in
          trapezoids := (!trapezoid_id, loop i) :: !trapezoids;
        end
      else
        begin
          (* De la droite vers la gauche. *) 
          let rec loop l =
            if l > j then
              []
            else
              (if l = i then
                (l,k)
              else
                if l = j then
                  (l,k+n)
                else
                  (l,1)) :: (loop (l + 1))
          in
          trapezoids := (!trapezoid_id, loop i) :: !trapezoids
        end;
      incr trapezoid_id
    end;
  (* Convert to integer programming problem. *)
  let clauses = Array.make n [] in
  List.iter
    begin fun (trapezoid_id, coefs) ->
      List.iter
        begin fun (i,ci) ->
          clauses.(i) <- (trapezoid_id, ci) :: clauses.(i)
        end
        coefs
    end
    !trapezoids;
  let remaining i = plonks.(i) - price.(i) in (* XXX *)
  let clauses =
    Array.mapi
      begin fun i cl ->
        (remaining i, cl)
      end
      clauses
  in
  let solution = Intprog.solve clauses in
  Intprog.check stdout clauses solution;
  Intprog.dump stdout solution
;;

let solve () =
  let ((marble_to_destination,plonks_to_do) as pb) = load !Opt.problem in
  let pb = { pb_sigma = marble_to_destination;
             pb_rho = plonks_to_do } in
  let solvers =
    let module S1 = Solver(G1) in
    (*let module S2 = Solver(G2) in
    let module S3 = Solver(G3) in*)
    [new S1.solver ~debug:!Opt.debug pb;
     (*new S2.solver ~debug:!Opt.debug pb;
     new S3.solver ~debug:!Opt.debug pb*)]
  in
  let counter = ref 0 in
  let not_found = ref true in
  try
    while true do
      List.iter
        begin fun s ->
          if !counter land 1023 = 0 then 
            begin
              Printf.printf "\r%d iterations (%s)%!" !counter s#get_status
            end;
          incr counter;
          match s#step with
          | None -> ()
          | Some l ->
              Format.printf "@.";
              Printf.printf "\nFound a solution:\n%!";
              let m = Array.length marble_to_destination in
              let result = Array.of_list (List.map (put_x m) (List.rev l)) in
              print_dance result;
              raise Found
        end
        solvers
    done
  with
  | Not_found -> Printf.printf "\nNo solution found.\n%!"
  | Found -> ()
;;

let _ =
  Arg.parse
    (Arg.align
      [ "-problem",     Arg.Set_string Opt.problem, "<file> set problem file";
        "-solve",       Arg.Unit solve,             " solve problem";
        "-residuals",   Arg.Unit residuals,         " print residuals";
        "-trapezoids",  Arg.Unit trapezoids,         " print trapezoids";
        "-simple",      Arg.Unit simple,            " solve problem simply";
        "-intprog",     Arg.Unit intprogify,        " solve problem using intprog";
        "-gui",         Arg.Unit gui,               " run in GUI";
        "-debug",       Arg.Set Opt.debug,          " enable debug mode" ])
    (fun x -> Opt.problem := x; gui ())
    "Bad args and stuff."
;;
