
(* 

OUTPUT_CHAR is at pos: 22235
  
*)

module LongMap = Map.Make(struct
      type t = int64
      let compare = compare
    end)

type instr =
| CMOVE of int * int * int
| GETMEM of int * int * int
| SETMEM of int * int * int
| ADD of int * int * int
| MUL of int * int * int
| DIV of int * int * int
| NOTAND of int * int * int
| EXIT
| ALLOC of int * int
| FREE of int
| OUTPUT of int
| INPUT of int
| GOTO of int * int
| CONST of int * int64
| NONE of int
| JUMP of int
| JUMPCONST of int
| GETMEMCONST of int * int * int
| ALLOCCONST of int * int
| BLOC of instr array
| SETCONSTMEM of int * int * int
| ADDCONST of int * int * int64
| NOTANDCONST of int * int * int64
| MULCONST of int * int * int64
| DIVCONST of int * int * int64
  
type values =
| Integer of int64
| Array of values_array
| Instr of int64 * instr

and values_array = {
    mutable array_dirty : bool;
    mutable array_tab : values array;
    mutable array_num : int64;
    mutable array_free : bool;
  }

let dump_file = "dump.bin"
let load_dump = ref None

let save_output = ref false
let saved_output = open_out "output.saved"

let narrays = ref 2L
let arrays = ref LongMap.empty

let buffered_output = Buffer.create 10000
  
let copy_output = ref None
  
let _ =
  Arg.parse [
    "-dump", Arg.String (fun s -> load_dump := Some s), " <dump_file> : load";
  ] (fun _ -> assert false) ""

let maxL = 0x100000000L
let minL = Int64.sub maxL 1L

let rec string_of_instr instr =
  match instr with
  | CMOVE (regA, regB, regC) ->
      Printf.sprintf "if not r%d then r%d := r%d" regC regA regB
  | GETMEM (regA, regB, regC) ->
      Printf.sprintf "r%d := r%d[r%d]" regA regB regC
  | SETMEM (regA, regB, regC) ->
      Printf.sprintf "r%d[r%d] <- r%d" regA regB regC
  | ADD (regA, regB, regC) ->
      Printf.sprintf "r%d := ADD(r%d,r%d)" regA regB regC 
      | MUL (regA, regB, regC) ->
      Printf.sprintf "r%d := MUL(r%d,r%d)" regA regB regC
  | DIV (regA, regB, regC) ->
      Printf.sprintf "r%d := DIV(r%d,r%d)" regA regB regC
  | NOTAND (regA, regB, regC) ->
      Printf.sprintf "r%d := NOTAND(r%d,r%d)" regA regB regC
  | EXIT -> 
      "EXIT"
  | ALLOC (regB, regC) -> 
      Printf.sprintf "r%d := ALLOC r%d" regB regC
  | FREE regC -> 
      Printf.sprintf "FREE r%d" regC
  | OUTPUT regC ->
      Printf.sprintf "OUTPUT r%d" regC
  | INPUT regC ->
      Printf.sprintf "r%d := INPUT" regC
  | GOTO (regB, regC) ->
      Printf.sprintf "GOTO r%d[r%d]" regB regC
  | CONST (regA, v) ->
      Printf.sprintf "r%d := %Ld %s" regA v
        (if v > 31L && v < 127L then 
          Printf.sprintf "'%c'" (char_of_int (Int64.to_int v))
        else "")
  | NONE operator -> 
      Printf.sprintf "NONE %d" operator
  | JUMP regC ->
      Printf.sprintf "JUMP r%d" regC
  | JUMPCONST n ->
      Printf.sprintf "JUMP %d\tJUMPCONST" n
  | GETMEMCONST (regA, regB, n) ->
      Printf.sprintf "r%d = r%d[%d]\t GETMEMCONST" regA regB n
  | ALLOCCONST (regB, n) ->
      Printf.sprintf "r%d = ALLOC %d\t ALLOCCONST" regB n
  | BLOC bloc ->
      let b = Buffer.create 1000 in
      Printf.bprintf b "\t{\n";
      for i = 0 to Array.length bloc - 1 do
        Printf.bprintf b "\t\t%s\n" (string_of_instr bloc.(i))
      done;
      Printf.bprintf b "\t}\n";
      Buffer.contents b
  | SETCONSTMEM (regA, n, regB) ->
      Printf.sprintf "r%d[%d] <- r%d\t SETCONSTMEM" regA n regB
  | ADDCONST (regA, regB, n) ->
      Printf.sprintf "r%d := ADD(r%d,%Ld)\t ADDCONST" regA regB n
  | NOTANDCONST (regA, regB, n) ->
      Printf.sprintf "r%d := NOTAND(r%d,%Ld)\t NOTANDCONST" regA regB n
  | MULCONST (regA, regB, n) ->
      Printf.sprintf "r%d := MUL(r%d,%Ld)\t MULCONST" regA regB n
  | DIVCONST (regA, regB, n) ->
      Printf.sprintf "r%d := DIV(r%d,%Ld)\t DIVCONST" regA regB n

      
      
let null = Integer 0L
  
let regs = Array.create 8 null

let get_int32 s pos =
  let pos4 = pos * 4 in
  let charA = int_of_char s.[pos4] in
  let charB = int_of_char s.[pos4+1] in
  let charC = int_of_char s.[pos4+2] in
  let charD = int_of_char s.[pos4+3] in
  let v =
    (charB lsl 16) lor
      (charC lsl 8) lor
      charD in
  Int64.logor (Int64.of_int v)  (Int64.shift_left (Int64.of_int charA)  24)

let mod32 v =  
  if v > minL || v < 0L then
    Int64.logand v minL
  else v
  
let integer v =
  Integer (mod32 v)
let program = 
  let tab =
    match !load_dump with
      None ->
        let ic = open_in "program" in
        let len = in_channel_length ic / 4 in
        let program = Array.create len null in
        let s = "    " in
        let rec iter i =
          if i < len then
            let nread = input ic s 0 4 in
            assert (nread = 4);
            program.(i) <- integer (get_int32 s 0);
            iter (i+1)
        in
        iter 0;
        program
    | Some _ -> [||]
  in
  {
    array_tab = tab;
    array_dirty = false;
    array_num = 0L;
    array_free = false;   
  }
  
let string_of_regs oc pos =
  Printf.fprintf oc "%d\t:" pos;
  for i = 0 to 7 do
    match regs.(i) with
      Integer v ->
        Printf.fprintf oc "%d=%4Ld " i v;
    | Array a ->
        Printf.fprintf oc "%d=[%d] " i (Array.length a.array_tab);
    | _ -> 
        Printf.fprintf oc "%d=???? " i;
  done;
  Printf.fprintf oc "\n"

let mem_get_const array pos =
  match array with
    Integer 0L | Instr (0L,_) ->
      program.array_tab.(pos)
  | Array a ->
      a.array_tab.(pos)
  | Integer _
  | Instr _ -> assert false
      
let mem_get array pos =
  let pos = match pos with
      Integer n -> Int64.to_int n
    | _ -> assert false
  in
  mem_get_const array pos     
      
let rewritten = ref false
  
let check_old_value v =
  if not !rewritten then
  match v with
    Instr (_,_) ->
      Printf.fprintf stderr "Rewritting code !\n";
      rewritten := true
  | _ -> ()
      
let mem_const_set array pos v =
  match array with
    Integer 0L | Instr (0L, _) ->
      if program.array_dirty then begin
          program.array_dirty <- false;
          program.array_tab <- Array.copy program.array_tab;
        end;
      check_old_value program.array_tab.(pos);
      program.array_tab.(pos) <- v
  
  | Array a ->
      if a.array_dirty then begin
          a.array_dirty <- false;
          a.array_tab <- Array.copy a.array_tab;
        end;      
      check_old_value a.array_tab.(pos);
      a.array_tab.(pos) <- v
  | Integer _
  | Instr _ -> assert false

let mem_set array pos v =
  let pos = match pos with
      Integer n -> Int64.to_int n
    | _ -> assert false
  in
  mem_const_set array pos v
  
let int32_of_value v =
  match v with
    Integer v | Instr (v,_) -> v
  | _ -> assert false
      
let int64_of_value v =
  match v with
    Integer v | Instr (v,_) ->  v
  | Array a -> 
      arrays := LongMap.add a.array_num a !arrays;
      !narrays

let int_of_value v =
  match v with
    Integer v | Instr (v,_) -> 
      Int64.to_int v 
  | _ -> assert false

let ninstrs = ref 0
let must_save = ref false

let ntranslates = ref 0

(*
let combine_instrs program pos v instr0 instr1 =
  match instr0, instr1 with
  | CONST (r0, n), GETMEM (ra, rb, rc) when rc = r0 && ra = r0 ->
      program.(pos) <- Instr (v, GETMEMCONST (ra, rb, n))
  | CONST (r0, n), ALLOC (rb, rc) when rb = r0 && rc = r0 ->
      program.(pos) <- Instr (v, ALLOCCONST (rb, n))
  | _ -> ()
*)      
  
let translate_instr program i =
  match program.(i) with
    Integer command ->
      let instr =     
        let operator = (Int64.to_int (Int64.shift_right command 28)) land 0xf in
        let arg = (Int64.to_int command) land 0xfffffff in
        let regA = (arg lsr 6) land 7 in
        let regB = (arg lsr 3) land 7 in
        let regC = arg land 7 in
        match operator with
          0 ->
            CMOVE (regA, regB, regC)
        | 1 -> 
            GETMEM (regA, regB, regC)
        | 2 ->
            SETMEM (regA, regB, regC)
        | 3 ->
            ADD (regA, regB, regC)
        | 4 ->
            MUL (regA, regB, regC)
        | 5 ->
            DIV (regA, regB, regC)
        | 6 ->
            NOTAND (regA, regB, regC)
        | 7 -> 
            EXIT
        | 8 -> 
            ALLOC (regB, regC)
        | 9 -> FREE regC
        | 10 -> OUTPUT regC
        | 11 -> INPUT regC
        | 12 -> 
            if regB = 6 then JUMP regC else GOTO (regB, regC)
        | 13 -> 
            let regA = (arg lsr 25) land 0x7 in
            let v = arg land 0x1ffffff in
            CONST (regA, Int64.of_int v)
        | _ -> NONE operator
      in
      program.(i) <- Instr (command, instr);
      (*
      if i > 0 then begin
          match program.(i-1) with
            Instr (v, instr0) ->
              combine_instrs program (i-1) v instr0 instr
          | _ -> ()
      end; *)
      instr
  | Instr (_, instr) -> instr
  | _ -> assert false
      
(*
let translate () =
  if !dump_translates then begin
      incr ntranslates;
      let oc = open_out (Printf.sprintf "program_%d.txt" !ntranslates) in
      let program = !program in
      for i = 0 to Array.length program - 1 do
        let instr = translate_instr program i in
        Printf.fprintf oc "%d\t: %s\n" i (string_of_instr instr);
      done;
      close_out oc
    end else begin
      let program = !program in
      for i = 0 to Array.length program - 1 do
        translate_instr program i
      done;
      
    end
      *)

let dump_to_file filename pos =
  Printf.fprintf stderr "dumping state to %s\n%!" filename;
  let oc = open_out filename  in
  Printf.fprintf stderr "(1)\n%!";
  output_value oc (program, regs, pos, Buffer.contents buffered_output);
  Printf.fprintf stderr "(2)\n%!";
  close_out oc;
  Printf.fprintf stderr "state dumped\n%!"

let dumptxt_to_file filename pos =
  let oc = open_out filename in
  let todump_map = ref LongMap.empty in
  let todump_fifo = Fifo.create () in
  let dumptxt_value v =
    match v with
      Integer v -> Printf.sprintf "%Ld" v
    | Instr (v, i) ->
        Printf.sprintf "%8Ld\t%s" v (string_of_instr i)
    | Array a ->
        if not a.array_free &&
          not (LongMap.mem a.array_num !todump_map) then begin
            Fifo.put todump_fifo a;
            todump_map := LongMap.add a.array_num a !todump_map;
          end;
        Printf.sprintf "Array %Ld" a.array_num;
  
  in
  Printf.fprintf oc "At pos: %d\n" pos;
  for i = 0 to 7 do
    Printf.fprintf oc "reg %d : %s\n" i (dumptxt_value regs.(i));
  done;
  
  Printf.fprintf oc "PROGRAM:\n";
  for i = 0 to Array.length program.array_tab - 1 do    
    Printf.fprintf oc "\t%d\t: %s\n" i (dumptxt_value program.array_tab.(i))
  done;

  todump_map := LongMap.add 0L program !todump_map;
  while Fifo.length todump_fifo > 0 do
    let a = Fifo.take todump_fifo in
    Printf.fprintf oc "Array %Ld:\n" a.array_num;
    for i = 0 to Array.length a.array_tab - 1 do
      Printf.fprintf oc "\t%d\t: %s\n" i (dumptxt_value a.array_tab.(i))
    done;
  done;
  close_out oc

let all_deads = 
  IntSet.add 0 (
    IntSet.add 1 (
      IntSet.add 2 (
        IntSet.add 3 (
          IntSet.add 4 (
            IntSet.add 5 (
              IntSet.add 6 (
                IntSet.add 7 IntSet.empty)))))))
  
let output_one_char c =
  begin
    match !copy_output with
      None -> ()
    | Some oc ->
        output_char oc c
  end;
  Buffer.add_char buffered_output c;
  output_char stdout c;
  flush stdout
  
let rec eval_code pos =
  let command = program.array_tab.(pos) in
  match command with
    Integer command -> 
      let instr = translate_instr program.array_tab pos in
      eval_instr pos instr eval_code
  | Array _ -> assert false
  | Instr (_, instr) -> 
      eval_instr pos instr eval_code

and eval_instr pos instr eval =
  
  match instr with
  | CMOVE (regA, regB, regC) ->
      (if regs.(regC) <> Integer 0L then regs.(regA) <- regs.(regB));
      eval (pos+1)
  | GETMEM (regA, regB, regC) ->
      regs.(regA) <- mem_get regs.(regB) regs.(regC);
      eval (pos+1)
  | SETMEM (regA, regB, regC) ->
      mem_set regs.(regA) regs.(regB) regs.(regC);
      eval (pos+1)
  | SETCONSTMEM (regA, n, regB) ->
      mem_const_set regs.(regA) n regs.(regB);
      eval (pos+1)
  | ADD (regA, regB, regC) ->
      let b = int64_of_value regs.(regB) in
      let c = int64_of_value regs.(regC) in
      let a = Int64.add b c in
      regs.(regA) <- integer a;
      eval (pos+1)
  | MUL (regA, regB, regC) ->
      let b = int64_of_value regs.(regB) in
      let c = int64_of_value regs.(regC) in
      let a = Int64.mul b c in
      regs.(regA) <- integer a;
      eval (pos+1)
  | DIV (regA, regB, regC) ->
      let b = int64_of_value regs.(regB) in
      let c = int64_of_value regs.(regC) in
      let a = Int64.div b c in
      regs.(regA) <- integer a;
      eval (pos+1)
  | NOTAND (regA, regB, regC) ->
      let b = int64_of_value regs.(regB) in
      let c = int64_of_value regs.(regC) in
      let a = Int64.lognot (Int64.logand b c) in
      let a = Int64.logand a minL in
      regs.(regA) <- integer a;
      eval (pos+1)
  | EXIT -> 
      close_out saved_output;
      exit 0
  | ALLOC (regB, regC) -> 
      narrays := Int64.add !narrays 1L;
      let a = {
          array_tab = Array.create (int_of_value regs.(regC)) null;
          array_num = !narrays;
          array_dirty = false;
          array_free = false;
        } in
      regs.(regB) <- Array a;
      eval (pos+1)
  | FREE regC -> 
      begin
        match regs.(regC) with
          Integer n | Instr (n,_) ->
            begin
              try
                let a = LongMap.find n !arrays in
                a.array_free <- true;
                a.array_tab <- [||];
                a.array_num <- 1L;
              with _ -> ()
            end;
            arrays := LongMap.remove n !arrays
        | Array a ->
            arrays := LongMap.remove a.array_num !arrays;
            a.array_free <- true;
            a.array_tab <- [||];
            a.array_num <- 1L;
      end;
      eval (pos+1)
  | OUTPUT regC ->
      let c = char_of_int (int_of_value regs.(regC)) in
      output_one_char c;
      eval (pos+1)
  | INPUT regC ->
      Printf.fprintf stderr ">%!";
      let c = input_char stdin in
      output_one_char c;
      if c = '!' then interpreter pos eval else begin
          regs.(regC) <- Integer (Int64.of_int (int_of_char c));
          eval (pos+1)
        end
  | GOTO (regB, regC) ->
      begin
        match regs.(regB) with
          Integer 0L | Instr (0L, _) -> ()
        | Array a -> 
            program.array_tab <- a.array_tab;
            program.array_dirty <- true;
            a.array_dirty <- true;
        | Integer _ | Instr _ -> assert false
      end;
      eval (int_of_value regs.(regC))
  | CONST (regA, v) ->
      regs.(regA) <- Integer v;
      eval (pos+1)
  | NONE operator -> 
      Printf.fprintf stderr "Unknown opcode %d\n" operator;
      exit 2
  | JUMP regC ->
      let new_pos = int_of_value regs.(regC) in
      begin
        match program.array_tab.(new_pos) with
          Instr(_, BLOC bloc) -> eval_bloc bloc 0
        | Instr(v, _) -> optimize_bloc v new_pos
        | _ -> eval_code new_pos
      end
  | JUMPCONST new_pos ->
      begin
        match program.array_tab.(new_pos) with
          Instr(_, BLOC bloc) -> eval_bloc bloc 0
        | Instr(v, _) -> optimize_bloc v new_pos
        | _ -> eval_code new_pos
      end
  | GETMEMCONST (regA, regB, n) ->
      regs.(regA) <- mem_get_const regs.(regB) n;
      eval (pos+1)
  | ALLOCCONST (regB, n) ->
      narrays := Int64.add !narrays 1L;
      let a = {
          array_tab = Array.create n null;
          array_num = !narrays;
          array_dirty = false;
          array_free = false;
        } in
      regs.(regB) <- Array a;
      eval (pos+1)
  | BLOC bloc ->
      eval_bloc bloc 0
  | ADDCONST (regA, regB, c) ->
      let b = int64_of_value regs.(regB) in
      let a = Int64.add b c in
      regs.(regA) <- integer a;
      eval (pos+1)
  | NOTANDCONST (regA, regB, c) ->
      let b = int64_of_value regs.(regB) in
      let a = Int64.lognot (Int64.logand b c) in
      let a = Int64.logand a minL in
      regs.(regA) <- integer a;
      eval (pos+1)
  | MULCONST (regA, regB, c) ->
      let b = int64_of_value regs.(regB) in
      let a = Int64.mul b c in
      regs.(regA) <- integer a;
      eval (pos+1)
  | DIVCONST (regA, regB, c) ->
      let b = int64_of_value regs.(regB) in
      let a = Int64.div b c in
      regs.(regA) <- integer a;
      eval (pos+1)

and eval_bloc bloc pos = 
  eval_instr pos bloc.(pos) (eval_bloc bloc)

and optimize_bloc v pos = 
  let rec iter pos instrs approx =
    match program.array_tab.(pos) with
      Instr (_, instr) ->
        begin
          match instr with
            JUMP reg ->
              let instr = 
                try
                  JUMPCONST (Int64.to_int (IntMap.find reg approx))
                with Not_found -> instr
              in
              instr :: instrs
          | EXIT 
          | NONE _
          | GOTO _ ->
              instr :: instrs
          | BLOC _ ->
              (JUMPCONST pos) :: instrs
          | CONST (reg, n) ->
              iter (pos+1) (instr :: instrs) (IntMap.add reg n approx)
          | GETMEM (regA, regB, regC) ->
              let instr =
                try
                  GETMEMCONST (regA, regB, Int64.to_int (IntMap.find regC approx))
                with Not_found -> instr
              in
              iter (pos+1) (instr :: instrs) (IntMap.remove regA approx)
          | ALLOC (regB, regC) ->
              let instr =
                try
                  ALLOCCONST (regB, Int64.to_int (IntMap.find regC approx))
                with Not_found -> instr
              in
              iter (pos+1) (instr :: instrs) (IntMap.remove regB approx)
          
          | SETMEM (regA, regB, regC) -> 
              let instr =
                try
                  SETCONSTMEM (regA, Int64.to_int (IntMap.find regB approx), regC)
                with Not_found -> instr
              in
              iter (pos+1) (instr :: instrs) approx
          
          | ADD (regA, regB, regC) ->
              let instr, approx =
                try
                  let nC = IntMap.find regC approx in
                  try
                    let a = mod32 (Int64.add nC (IntMap.find regB approx)) in
                    CONST (regA, a), IntMap.add regA a approx
                  with Not_found ->
                      (ADDCONST (regA, regB, nC)), IntMap.remove regA approx
                with Not_found -> 
                    (try
                        ADDCONST (regA, regC, IntMap.find regB approx)
                      with Not_found -> instr), IntMap.remove regA approx
              in              
              iter (pos+1) (instr :: instrs) approx
          
          | MUL (regA, regB, regC) ->
              let instr, approx =
                try
                  let nC = IntMap.find regC approx in
                  try
                    let a = mod32 (Int64.mul nC (IntMap.find regB approx)) in
                    CONST (regA, a), IntMap.add regA a approx
                  with Not_found ->
                      (MULCONST (regA, regB, nC)), IntMap.remove regA approx
                with Not_found -> 
                    (try
                        MULCONST (regA, regC, IntMap.find regB approx)
                      with Not_found -> instr), IntMap.remove regA approx
              in              
              iter (pos+1) (instr :: instrs) approx
          
          | DIV (regA, regB, regC) ->
              let instr, approx =
                try
                  let nC = IntMap.find regC approx in
                  try
                    let a = mod32 (Int64.div nC (IntMap.find regB approx)) in
                    CONST (regA, a), IntMap.add regA a approx
                  with Not_found ->
                      (DIVCONST (regA, regB, nC)), IntMap.remove regA approx
                with Not_found -> 
                    instr, IntMap.remove regA approx
              in              
              iter (pos+1) (instr :: instrs) approx
              
          | NOTAND (regA, regB, regC) ->
              let instr, approx =
                try
                  let nC = IntMap.find regC approx in
                  try
                    let a = mod32 (Int64.lognot (Int64.logand nC (IntMap.find regB approx))) in
                    CONST (regA, a), IntMap.add regA a approx
                  with Not_found ->
                      NOTANDCONST (regA, regB, nC), IntMap.remove regA approx
                with Not_found -> 
                    (try
                      NOTANDCONST (regA, regC, IntMap.find regB approx)
                    with Not_found -> instr), IntMap.remove regA approx                    
              in              
              iter (pos+1) (instr :: instrs) approx
              
          | _ ->
              let approx = 
                match instr with
                | JUMPCONST _
                | EXIT
                | NONE _
                | GOTO _
                | JUMP _ 
                | BLOC _
                  -> assert false  
                | OUTPUT _
                | FREE _
                | SETMEM _
                | SETCONSTMEM _
                  -> approx
                | ALLOCCONST (reg, _) 
                | GETMEMCONST (reg, _, _)
                | CONST (reg, _)
                | INPUT reg
                | ALLOC (reg, _)
                | NOTAND (reg, _, _)
                | DIV (reg, _, _)
                | MUL (reg, _, _)
                | ADD (reg, _, _)
                | ADDCONST (reg, _, _)
                | MULCONST (reg, _, _)
                | DIVCONST (reg, _, _)
                | NOTANDCONST (reg, _, _)
                | GETMEM (reg, _, _)
                | CMOVE (reg, _, _)
                  -> IntMap.remove reg approx                  
              in
              iter (pos+1) (instr :: instrs) approx
        end
    | Integer v ->
        ignore (translate_instr program.array_tab pos);
        iter pos instrs approx
    | _ -> assert false
  in
  let bloc = iter pos [] (IntMap.add 6 0L IntMap.empty) in
    
  let rec remove_dead instrs deads rems =
    match instrs with
      [] -> rems
    | instr :: tail ->
        let rems =
          match instr with
          | CONST (reg, _) ->
              if IntSet.mem reg deads then begin
                  rems
                end
              else
                instr :: rems
          | _ -> instr :: rems
        in
        let deads = 
          match instr with
          | EXIT
          | NONE _ -> all_deads
          | CONST (reg,_)
          | ALLOCCONST (reg,_)
          | INPUT reg
            -> 
              IntSet.add reg deads

          | GETMEMCONST (regA, regB, _) 
          | ALLOC (regA, regB)
          | ADDCONST (regA, regB, _) 
          | MULCONST (regA, regB, _) 
          | DIVCONST (regA, regB, _) 
          | NOTANDCONST (regA, regB, _) 
            -> IntSet.remove regB (IntSet.add regA deads)
              
          | OUTPUT reg
          | FREE reg
            -> IntSet.remove reg deads
             
          | SETMEM (regA, regB, regC) ->
              IntSet.remove regA (IntSet.remove regB (IntSet.remove regC deads))              
          | SETCONSTMEM (regA, n, regB) ->
              IntSet.remove regA (IntSet.remove regB deads)


          | CMOVE (regA, regB, regC) ->
              IntSet.remove regB (IntSet.remove regA deads)
            
          | GETMEM (regA, regB, regC) 
          | NOTAND (regA, regB, regC) 
          | ADD (regA, regB, regC) 
          | MUL (regA, regB, regC) 
          | DIV (regA, regB, regC) 
            -> IntSet.remove regC (IntSet.remove regB (IntSet.add regA deads))

          | JUMPCONST _
          | GOTO _
          | JUMP _ 
          | BLOC _
            -> IntSet.empty
        in
        remove_dead tail deads rems
  in
  let bloc = remove_dead bloc IntSet.empty [] in
    
  let bloc = Array.of_list bloc in
  program.array_tab.(pos) <- Instr(v, BLOC bloc);
  eval_code pos

                  
                
  
and interpreter pos eval =
  Printf.fprintf stderr "DEBUG:%!";
  let command = input_line stdin in
  let s = String2.split_simplify command ' ' in
  match s with
    [] -> interpreter pos eval
  | [ "exit" ] -> eval pos
  | [ "regs" ] -> 
      string_of_regs stderr pos;
      interpreter pos eval
  | ["dumpbin"; filename ] ->
      dump_to_file filename pos;
      interpreter pos eval
  | ["dumptxt"; filename ] ->
      dumptxt_to_file filename pos;
      interpreter pos eval
  | ["output"; filename] ->
      begin
        match !copy_output with
          None -> ()
        | Some oc -> close_out oc
      end;
      copy_output := Some (open_out filename);
      interpreter pos eval
      
  | _ -> 
      Printf.fprintf stderr "BAD COMMAND: %s\nUse 'exit' to exit\n" command;
      interpreter pos eval


let _ = 
  ignore (Sys.signal Sys.sighup (Sys.Signal_handle (fun _ ->
          Printf.fprintf stderr "must save...\n";
          must_save := true
    )));
  match !load_dump with
    Some dump_file ->
      let ic = open_in dump_file in
      let (program_saved, regs_saved, pos_saved, before) = input_value ic in
      close_in ic;
      program.array_tab <- program_saved.array_tab;
      program.array_dirty <- program_saved.array_dirty;
      
      for i = 0 to 7 do
        regs.(i) <- regs_saved.(i);
      done;
      ninstrs := 1;
      Printf.fprintf stderr "Restarting from state\n" ;
      Buffer.add_string buffered_output before;
      Printf.fprintf stderr "%s" before;
      eval_code pos_saved
  | None ->
      eval_code 0
  
  
  

  